from veros import runtime_settings


class Variable:
    def __init__(
        self,
        name,
        dims,
        units="",
        long_description="",
        dtype=None,
        time_dependent=True,
        scale=1.0,
        write_to_restart=False,
        extra_attributes=None,
        mask=None,
        active=True,
        initial=None,
    ):
        self.name = name
        self.dims = dims
        self.units = units
        self.long_description = long_description
        self.dtype = dtype
        self.time_dependent = time_dependent
        self.scale = scale
        self.write_to_restart = write_to_restart
        self.active = active
        self.initial = initial

        self.get_mask = lambda settings, vs: None

        if mask is not None:
            if not callable(mask):
                raise TypeError("mask argument has to be callable")

            self.get_mask = mask

        elif isinstance(dims, tuple):
            if dims[:3] in DEFAULT_MASKS:
                self.get_mask = DEFAULT_MASKS[dims[:3]]
            elif dims[:2] in DEFAULT_MASKS:
                self.get_mask = DEFAULT_MASKS[dims[:2]]

        #: Additional attributes to be written in netCDF output
        self.extra_attributes = extra_attributes or {}

    def __repr__(self):
        attr_str = []
        for v in vars(self):
            attr_str.append(f"{v}={getattr(self, v)}")
        attr_str = ", ".join(attr_str)
        return f"{self.__class__.__qualname__}({attr_str})"


# fill value for netCDF output (invalid floating data is replaced by this value)
FLOAT_FILL_VALUE = -1e18

#
XT = ("xt",)
XU = ("xu",)
YT = ("yt",)
YU = ("yu",)
ZT = ("zt",)
ZW = ("zw",)
T_HOR = ("xt", "yt")
U_HOR = ("xu", "yt")
V_HOR = ("xt", "yu")
ZETA_HOR = ("xu", "yu")
T_GRID = ("xt", "yt", "zt")
U_GRID = ("xu", "yt", "zt")
V_GRID = ("xt", "yu", "zt")
W_GRID = ("xt", "yt", "zw")
ZETA_GRID = ("xu", "yu", "zt")
TIMESTEPS = ("timesteps",)
ISLE = ("isle",)
TENSOR_COMP = ("tensor1", "tensor2")

# those are written to netCDF output by default
BASE_DIMENSIONS = XT + XU + YT + YU + ZT + ZW + ISLE
GHOST_DIMENSIONS = ("xt", "yt", "xu", "yu")

# these are the settings that are getting used to determine shapes
DIM_TO_SHAPE_VAR = {
    "xt": "nx",
    "xu": "nx",
    "yt": "ny",
    "yu": "ny",
    "zt": "nz",
    "zw": "nz",
    "timesteps": 3,
    "tensor1": 2,
    "tensor2": 2,
    "isle": 0,
}

DEFAULT_MASKS = {
    T_HOR: lambda settings, vs: vs.maskT[:, :, -1],
    U_HOR: lambda settings, vs: vs.maskU[:, :, -1],
    V_HOR: lambda settings, vs: vs.maskV[:, :, -1],
    ZETA_HOR: lambda settings, vs: vs.maskZ[:, :, -1],
    T_GRID: lambda settings, vs: vs.maskT,
    U_GRID: lambda settings, vs: vs.maskU,
    V_GRID: lambda settings, vs: vs.maskV,
    W_GRID: lambda settings, vs: vs.maskW,
    ZETA_GRID: lambda settings, vs: vs.maskZ,
}


# custom mask for streamfunction
def _get_psi_mask(settings, vs):
    if not settings.enable_streamfunction:
        return vs.maskT[:, :, -1]

    # eroded around the edges
    return vs.maskZ[:, :, -1] | ~vs.isle_boundary_mask


def get_fill_value(dtype):
    import numpy as onp

    if onp.issubdtype(dtype, onp.floating):
        return FLOAT_FILL_VALUE

    return onp.iinfo(dtype).max


def get_shape(dimensions, grid, include_ghosts=True, local=True):
    from veros.routines import CURRENT_CONTEXT
    from veros.distributed import SCATTERED_DIMENSIONS

    if grid is None:
        return ()

    px, py = runtime_settings.num_proc
    grid_shapes = dict(dimensions)

    if local and CURRENT_CONTEXT.is_dist_safe:
        for pxi, dims in zip((px, py), SCATTERED_DIMENSIONS):
            for dim in dims:
                if dim not in grid_shapes:
                    continue

                grid_shapes[dim] = grid_shapes[dim] // pxi

    if include_ghosts:
        for d in GHOST_DIMENSIONS:
            if d in grid_shapes:
                grid_shapes[d] += 4

    shape = []
    for grid_dim in grid:
        if isinstance(grid_dim, int):
            shape.append(grid_dim)
            continue

        if grid_dim not in grid_shapes:
            raise ValueError(f"unrecognized dimension {grid_dim}")

        shape.append(grid_shapes[grid_dim])

    return tuple(shape)


def remove_ghosts(array, dims):
    if dims is None:
        # scalar
        return array

    ghost_mask = tuple(slice(2, -2) if dim in GHOST_DIMENSIONS else slice(None) for dim in dims)
    return array[ghost_mask]


VARIABLES = {
    # scalars
    "tau": Variable(
        "Index of current time step",
        None,
        "",
        "Index of current time step",
        dtype="int32",
        initial=1,
        write_to_restart=True,
    ),
    "taup1": Variable(
        "Index of next time step", None, "", "Index of next time step", dtype="int32", initial=2, write_to_restart=True
    ),
    "taum1": Variable(
        "Index of last time step", None, "", "Index of last time step", dtype="int32", initial=0, write_to_restart=True
    ),
    "time": Variable(
        "Current time",
        None,
        "",
        "Current time",
        write_to_restart=True,
    ),
    "itt": Variable("Current iteration", None, "", "Current iteration", dtype="int32", initial=0),
    # base variables
    "dxt": Variable("Zonal T-grid spacing", XT, "m", "Zonal (x) spacing of T-grid point", time_dependent=False),
    "dxu": Variable("Zonal U-grid spacing", XU, "m", "Zonal (x) spacing of U-grid point", time_dependent=False),
    "dyt": Variable(
        "Meridional T-grid spacing", YT, "m", "Meridional (y) spacing of T-grid point", time_dependent=False
    ),
    "dyu": Variable(
        "Meridional U-grid spacing", YU, "m", "Meridional (y) spacing of U-grid point", time_dependent=False
    ),
    "zt": Variable(
        "Vertical coordinate (T)",
        ZT,
        "m",
        "Vertical coordinate",
        time_dependent=False,
        extra_attributes={"positive": "up"},
    ),
    "zw": Variable(
        "Vertical coordinate (W)",
        ZW,
        "m",
        "Vertical coordinate",
        time_dependent=False,
        extra_attributes={"positive": "up"},
    ),
    "dzt": Variable("Vertical spacing (T)", ZT, "m", "Vertical spacing", time_dependent=False),
    "dzw": Variable("Vertical spacing (W)", ZW, "m", "Vertical spacing", time_dependent=False),
    "cost": Variable("Metric factor (T)", YT, "1", "Metric factor for spherical coordinates", time_dependent=False),
    "cosu": Variable("Metric factor (U)", YU, "1", "Metric factor for spherical coordinates", time_dependent=False),
    "tantr": Variable("Metric factor", YT, "1", "Metric factor for spherical coordinates", time_dependent=False),
    "coriolis_t": Variable(
        "Coriolis frequency", T_HOR, "1/s", "Coriolis frequency at T grid point", time_dependent=False
    ),
    "kbot": Variable(
        "Index of deepest cell",
        T_HOR,
        "",
        "Index of the deepest grid cell (counting from 1, 0 means all land)",
        dtype="int32",
        time_dependent=False,
    ),
    "ht": Variable("Total depth (T)", T_HOR, "m", "Total depth of the water column", time_dependent=False),
    "hu": Variable("Total depth (U)", U_HOR, "m", "Total depth of the water column", time_dependent=False),
    "hv": Variable("Total depth (V)", V_HOR, "m", "Total depth of the water column", time_dependent=False),
    "hur": Variable(
        "Total depth (U), masked", U_HOR, "m", "Total depth of the water column (masked)", time_dependent=False
    ),
    "hvr": Variable(
        "Total depth (V), masked", V_HOR, "m", "Total depth of the water column (masked)", time_dependent=False
    ),
    "beta": Variable(
        "Change of Coriolis freq.", T_HOR, "1/(ms)", "Change of Coriolis frequency with latitude", time_dependent=False
    ),
    "area_t": Variable("Area of T-box", T_HOR, "m^2", "Area of T-box", time_dependent=False),
    "area_u": Variable("Area of U-box", U_HOR, "m^2", "Area of U-box", time_dependent=False),
    "area_v": Variable("Area of V-box", V_HOR, "m^2", "Area of V-box", time_dependent=False),
    "maskT": Variable(
        "Mask for tracer points",
        T_GRID,
        "",
        "Mask in physical space for tracer points",
        time_dependent=False,
        dtype="bool",
    ),
    "maskU": Variable(
        "Mask for U points",
        U_GRID,
        "",
        "Mask in physical space for U points",
        time_dependent=False,
        dtype="bool",
    ),
    "maskV": Variable(
        "Mask for V points",
        V_GRID,
        "",
        "Mask in physical space for V points",
        time_dependent=False,
        dtype="bool",
    ),
    "maskW": Variable(
        "Mask for W points",
        W_GRID,
        "",
        "Mask in physical space for W points",
        time_dependent=False,
        dtype="bool",
    ),
    "maskZ": Variable(
        "Mask for Zeta points",
        ZETA_GRID,
        "",
        "Mask in physical space for Zeta points",
        time_dependent=False,
        dtype="bool",
    ),
    "rho": Variable(
        "Density",
        T_GRID + TIMESTEPS,
        "kg/m^3",
        "In-situ density anomaly, relative to the surface mean value of 1024 kg/m^3",
        write_to_restart=True,
    ),
    "prho": Variable(
        "Potential density",
        T_GRID,
        "kg/m^3",
        "Potential density anomaly, relative to the surface mean value of 1024 kg/m^3 "
        "(identical to in-situ density anomaly for equation of state type 1, 2, and 4)",
    ),
    "int_drhodT": Variable(
        "Der. of dyn. enthalpy by temperature",
        T_GRID + TIMESTEPS,
        "kg / (m^2 deg C)",
        "Partial derivative of dynamic enthalpy by temperature",
        write_to_restart=True,
    ),
    "int_drhodS": Variable(
        "Der. of dyn. enthalpy by salinity",
        T_GRID + TIMESTEPS,
        "kg / (m^2 g / kg)",
        "Partial derivative of dynamic enthalpy by salinity",
        write_to_restart=True,
    ),
    "Nsqr": Variable(
        "Square of stability frequency",
        W_GRID + TIMESTEPS,
        "1/s^2",
        "Square of stability frequency",
        write_to_restart=True,
    ),
    "Hd": Variable("Dynamic enthalpy", T_GRID + TIMESTEPS, "m^2/s^2", "Dynamic enthalpy", write_to_restart=True),
    "dHd": Variable(
        "Change of dyn. enth. by adv.",
        T_GRID + TIMESTEPS,
        "m^2/s^3",
        "Change of dynamic enthalpy due to advection",
        write_to_restart=True,
    ),
    "temp": Variable("Temperature", T_GRID + TIMESTEPS, "deg C", "Conservative temperature", write_to_restart=True),
    "dtemp": Variable(
        "Temperature tendency",
        T_GRID + TIMESTEPS,
        "deg C/s",
        "Conservative temperature tendency",
        write_to_restart=True,
    ),
    "salt": Variable("Salinity", T_GRID + TIMESTEPS, "g/kg", "Salinity", write_to_restart=True),
    "dsalt": Variable("Salinity tendency", T_GRID + TIMESTEPS, "g/(kg s)", "Salinity tendency", write_to_restart=True),
    "dtemp_vmix": Variable(
        "Change of temp. by vertical mixing",
        T_GRID,
        "deg C/s",
        "Change of temperature due to vertical mixing",
    ),
    "dtemp_hmix": Variable(
        "Change of temp. by horizontal mixing",
        T_GRID,
        "deg C/s",
        "Change of temperature due to horizontal mixing",
    ),
    "dsalt_vmix": Variable(
        "Change of sal. by vertical mixing",
        T_GRID,
        "deg C/s",
        "Change of salinity due to vertical mixing",
    ),
    "dsalt_hmix": Variable(
        "Change of sal. by horizontal mixing",
        T_GRID,
        "deg C/s",
        "Change of salinity due to horizontal mixing",
    ),
    "dtemp_iso": Variable(
        "Change of temp. by isop. mixing",
        T_GRID,
        "deg C/s",
        "Change of temperature due to isopycnal mixing plus skew mixing",
    ),
    "dsalt_iso": Variable(
        "Change of sal. by isop. mixing",
        T_GRID,
        "deg C/s",
        "Change of salinity due to isopycnal mixing plus skew mixing",
    ),
    "forc_temp_surface": Variable(
        "Surface temperature flux",
        T_HOR,
        "m deg C/s",
        "Surface temperature flux",
    ),
    "forc_salt_surface": Variable(
        "Surface salinity flux",
        T_HOR,
        "m g/s kg",
        "Surface salinity flux",
    ),
    "u": Variable("Zonal velocity", U_GRID + TIMESTEPS, "m/s", "Zonal velocity", write_to_restart=True),
    "v": Variable("Meridional velocity", V_GRID + TIMESTEPS, "m/s", "Meridional velocity", write_to_restart=True),
    "w": Variable("Vertical velocity", W_GRID + TIMESTEPS, "m/s", "Vertical velocity", write_to_restart=True),
    "du": Variable(
        "Zonal velocity tendency", U_GRID + TIMESTEPS, "m/s", "Zonal velocity tendency", write_to_restart=True
    ),
    "dv": Variable(
        "Meridional velocity tendency", V_GRID + TIMESTEPS, "m/s", "Meridional velocity tendency", write_to_restart=True
    ),
    "du_cor": Variable("Change of u by Coriolis force", U_GRID, "m/s^2", "Change of u due to Coriolis force"),
    "dv_cor": Variable("Change of v by Coriolis force", V_GRID, "m/s^2", "Change of v due to Coriolis force"),
    "du_mix": Variable(
        "Change of u by vertical mixing", U_GRID, "m/s^2", "Change of u due to implicit vertical mixing"
    ),
    "dv_mix": Variable(
        "Change of v by vertical mixing", V_GRID, "m/s^2", "Change of v due to implicit vertical mixing"
    ),
    "du_adv": Variable("Change of u by advection", U_GRID, "m/s^2", "Change of u due to advection"),
    "dv_adv": Variable("Change of v by advection", V_GRID, "m/s^2", "Change of v due to advection"),
    "p_hydro": Variable("Hydrostatic pressure", T_GRID, "m^2/s^2", "Hydrostatic pressure"),
    "kappaM": Variable("Vertical viscosity", T_GRID, "m^2/s", "Vertical viscosity"),
    "kappaH": Variable("Vertical diffusivity", W_GRID, "m^2/s", "Vertical diffusivity"),
    "surface_taux": Variable(
        "Surface wind stress",
        U_HOR,
        "N/m^2",
        "Zonal surface wind stress",
    ),
    "surface_tauy": Variable(
        "Surface wind stress",
        V_HOR,
        "N/m^2",
        "Meridional surface wind stress",
    ),
    "forc_rho_surface": Variable("Surface density flux", T_HOR, "kg / (m^2 s)", "Surface potential density flux"),
    "ssh": Variable(
        "Sea surface height",
        T_HOR,
        "m",
        "Sea surface height",
        active=lambda settings: not settings.enable_streamfunction,
    ),
    "psi": Variable(
        lambda settings: "Streamfunction" if settings.enable_streamfunction else "Surface pressure",
        lambda settings: ZETA_HOR + TIMESTEPS if settings.enable_streamfunction else T_HOR + TIMESTEPS,
        lambda settings: "m^3/s" if settings.enable_streamfunction else "m^2/s^2",
        lambda settings: "Barotropic streamfunction" if settings.enable_streamfunction else "Surface pressure",
        write_to_restart=True,
        mask=_get_psi_mask,
    ),
    "dpsi": Variable(
        "Streamfunction tendency",
        ZETA_HOR + TIMESTEPS,
        "m^3/s^2",
        "Streamfunction tendency",
        write_to_restart=True,
        active=lambda settings: settings.enable_streamfunction,
    ),
    "land_map": Variable(
        "Land map",
        T_HOR,
        "",
        "Land map",
        dtype="int32",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "isle": Variable(
        "Island number",
        ISLE,
        "",
        "Island number",
        dtype="int32",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "psin": Variable(
        "Boundary streamfunction",
        ZETA_HOR + ISLE,
        "m^3/s",
        "Boundary streamfunction",
        time_dependent=False,
        mask=_get_psi_mask,
        active=lambda settings: settings.enable_streamfunction,
    ),
    "dpsin": Variable(
        "Boundary streamfunction factor",
        ISLE + TIMESTEPS,
        "m^3/s^2",
        "Boundary streamfunction factor",
        write_to_restart=True,
        active=lambda settings: settings.enable_streamfunction,
    ),
    "line_psin": Variable(
        "Boundary line integrals",
        ISLE + ISLE,
        "m^4/s^2",
        "Boundary line integrals",
        time_dependent=False,
        active=lambda settings: settings.enable_streamfunction,
    ),
    "isle_boundary_mask": Variable(
        "Island boundary mask",
        T_HOR,
        "",
        "Island boundary mask",
        time_dependent=False,
        dtype="bool",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "line_dir_south_mask": Variable(
        "Line integral mask",
        T_HOR + ISLE,
        "",
        "Line integral mask",
        time_dependent=False,
        dtype="bool",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "line_dir_north_mask": Variable(
        "Line integral mask",
        T_HOR + ISLE,
        "",
        "Line integral mask",
        time_dependent=False,
        dtype="bool",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "line_dir_east_mask": Variable(
        "Line integral mask",
        T_HOR + ISLE,
        "",
        "Line integral mask",
        time_dependent=False,
        dtype="bool",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "line_dir_west_mask": Variable(
        "Line integral mask",
        T_HOR + ISLE,
        "",
        "Line integral mask",
        time_dependent=False,
        dtype="bool",
        active=lambda settings: settings.enable_streamfunction,
    ),
    "K_gm": Variable("Skewness diffusivity", W_GRID, "m^2/s", "GM diffusivity, either constant or from EKE model"),
    "K_iso": Variable("Isopycnal diffusivity", W_GRID, "m^2/s", "Along-isopycnal diffusivity"),
    "K_diss_v": Variable(
        "Dissipation of kinetic Energy",
        W_GRID,
        "m^2/s^3",
        "Kinetic energy dissipation by vertical, rayleigh and bottom friction",
        write_to_restart=True,
    ),
    "K_diss_bot": Variable(
        "Dissipation of kinetic Energy", W_GRID, "m^2/s^3", "Mean energy dissipation by bottom and rayleigh friction"
    ),
    "K_diss_h": Variable(
        "Dissipation of kinetic Energy", W_GRID, "m^2/s^3", "Kinetic energy dissipation by horizontal friction"
    ),
    "K_diss_gm": Variable(
        "Dissipation of mean energy",
        W_GRID,
        "m^2/s^3",
        "Mean energy dissipation by GM (TRM formalism only)",
    ),
    "P_diss_v": Variable(
        "Dissipation of potential Energy", W_GRID, "m^2/s^3", "Potential energy dissipation by vertical diffusion"
    ),
    "P_diss_nonlin": Variable(
        "Dissipation of potential Energy",
        W_GRID,
        "m^2/s^3",
        "Potential energy dissipation by nonlinear equation of state",
    ),
    "P_diss_iso": Variable(
        "Dissipation of potential Energy", W_GRID, "m^2/s^3", "Potential energy dissipation by isopycnal mixing"
    ),
    "P_diss_skew": Variable(
        "Dissipation of potential Energy", W_GRID, "m^2/s^3", "Potential energy dissipation by GM (w/o TRM)"
    ),
    "P_diss_hmix": Variable(
        "Dissipation of potential Energy", W_GRID, "m^2/s^3", "Potential energy dissipation by horizontal mixing"
    ),
    "P_diss_adv": Variable(
        "Dissipation of potential Energy", W_GRID, "m^2/s^3", "Potential energy dissipation by advection"
    ),
    "P_diss_sources": Variable(
        "Dissipation of potential Energy",
        W_GRID,
        "m^2/s^3",
        "Potential energy dissipation by external sources (e.g. restoring zones)",
    ),
    "u_wgrid": Variable("U on W grid", W_GRID, "m/s", "Zonal velocity interpolated to W grid points"),
    "v_wgrid": Variable("V on W grid", W_GRID, "m/s", "Meridional velocity interpolated to W grid points"),
    "w_wgrid": Variable("W on W grid", W_GRID, "m/s", "Vertical velocity interpolated to W grid points"),
    "xt": Variable(
        "Zonal coordinate (T)",
        XT,
        lambda settings: "degrees_east" if settings.coord_degree else "km",
        "Zonal (x) coordinate of T-grid point",
        time_dependent=False,
        scale=lambda settings: 1 if settings.coord_degree else 1e-3,
    ),
    "xu": Variable(
        "Zonal coordinate (U)",
        XU,
        lambda settings: "degrees_east" if settings.coord_degree else "km",
        "Zonal (x) coordinate of U-grid point",
        time_dependent=False,
        scale=lambda settings: 1 if settings.coord_degree else 1e-3,
    ),
    "yt": Variable(
        "Meridional coordinate (T)",
        YT,
        lambda settings: "degrees_north" if settings.coord_degree else "km",
        "Meridional (y) coordinate of T-grid point",
        time_dependent=False,
        scale=lambda settings: 1 if settings.coord_degree else 1e-3,
    ),
    "yu": Variable(
        "Meridional coordinate (U)",
        YU,
        lambda settings: "degrees_north" if settings.coord_degree else "km",
        "Meridional (y) coordinate of U-grid point",
        time_dependent=False,
        scale=lambda settings: 1 if settings.coord_degree else 1e-3,
    ),
    "temp_source": Variable(
        "Source of temperature",
        T_GRID,
        "K/s",
        "Non-conservative source of temperature",
        active=lambda settings: settings.enable_tempsalt_sources,
    ),
    "salt_source": Variable(
        "Source of salt",
        T_GRID,
        "g/(kg s)",
        "Non-conservative source of salt",
        active=lambda settings: settings.enable_tempsalt_sources,
    ),
    "u_source": Variable(
        "Source of zonal velocity",
        U_GRID,
        "m/s^2",
        "Non-conservative source of zonal velocity",
        active=lambda settings: settings.enable_momentum_sources,
    ),
    "v_source": Variable(
        "Source of meridional velocity",
        V_GRID,
        "m/s^2",
        "Non-conservative source of meridional velocity",
        active=lambda settings: settings.enable_momentum_sources,
    ),
    "K_11": Variable(
        "Isopycnal mixing coefficient",
        T_GRID,
        "m^2/s",
        "Isopycnal mixing tensor component",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "K_22": Variable(
        "Isopycnal mixing coefficient",
        T_GRID,
        "m^2/s",
        "Isopycnal mixing tensor component",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "K_33": Variable(
        "Isopycnal mixing coefficient",
        T_GRID,
        "m^2/s",
        "Isopycnal mixing tensor component",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "Ai_ez": Variable(
        "Isopycnal diffusion coefficient",
        T_GRID + TENSOR_COMP,
        "Vertical isopycnal diffusion coefficient on eastern face of T cell",
        "1",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "Ai_nz": Variable(
        "Isopycnal diffusion coefficient",
        T_GRID + TENSOR_COMP,
        "Vertical isopycnal diffusion coefficient on northern face of T cell",
        "1",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "Ai_bx": Variable(
        "Isopycnal diffusion coefficient",
        T_GRID + TENSOR_COMP,
        "Zonal isopycnal diffusion coefficient on bottom face of T cell",
        "1",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "Ai_by": Variable(
        "Isopycnal diffusion coefficient",
        T_GRID + TENSOR_COMP,
        "Meridional isopycnal diffusion coefficient on bottom face of T cell",
        "1",
        active=lambda settings: settings.enable_neutral_diffusion,
    ),
    "B1_gm": Variable(
        "Zonal component of GM streamfunction",
        V_GRID,
        "m^2/s",
        "Zonal component of GM streamfunction",
        active=lambda settings: settings.enable_skew_diffusion,
    ),
    "B2_gm": Variable(
        "Meridional component of GM streamfunction",
        U_GRID,
        "m^2/s",
        "Meridional component of GM streamfunction",
        active=lambda settings: settings.enable_skew_diffusion,
    ),
    "r_bot_var_u": Variable(
        "Bottom friction coeff.",
        U_HOR,
        "1/s",
        "Zonal bottom friction coefficient",
        active=lambda settings: settings.enable_bottom_friction_var,
    ),
    "r_bot_var_v": Variable(
        "Bottom friction coeff.",
        V_HOR,
        "1/s",
        "Meridional bottom friction coefficient",
        active=lambda settings: settings.enable_bottom_friction_var,
    ),
    "kappa_gm": Variable(
        "Vertical diffusivity",
        W_GRID,
        "m^2/s",
        "Vertical diffusivity",
        active=lambda settings: settings.enable_TEM_friction,
    ),
    "tke": Variable(
        "Turbulent kinetic energy",
        W_GRID + TIMESTEPS,
        "m^2/s^2",
        "Turbulent kinetic energy",
        write_to_restart=True,
        active=lambda settings: settings.enable_tke,
    ),
    "sqrttke": Variable(
        "Square-root of TKE",
        W_GRID,
        "m/s",
        "Square-root of TKE",
        active=lambda settings: settings.enable_tke,
    ),
    "dtke": Variable(
        "Turbulent kinetic energy tendency",
        W_GRID + TIMESTEPS,
        "m^2/s^3",
        "Turbulent kinetic energy tendency",
        write_to_restart=True,
        active=lambda settings: settings.enable_tke,
    ),
    "Prandtlnumber": Variable(
        "Prandtl number",
        W_GRID,
        "",
        "Prandtl number",
        active=lambda settings: settings.enable_tke,
    ),
    "mxl": Variable(
        "Mixing length",
        W_GRID,
        "m",
        "Mixing length",
        active=lambda settings: settings.enable_tke,
    ),
    "forc_tke_surface": Variable(
        "TKE surface flux",
        T_HOR,
        "m^3/s^3",
        "TKE surface flux",
        active=lambda settings: settings.enable_tke,
    ),
    "tke_diss": Variable(
        "TKE dissipation",
        W_GRID,
        "m^2/s^3",
        "TKE dissipation",
        active=lambda settings: settings.enable_tke,
    ),
    "tke_surf_corr": Variable(
        "Correction of TKE surface flux",
        T_HOR,
        "m^3/s^3",
        "Correction of TKE surface flux",
        active=lambda settings: settings.enable_tke,
    ),
    "eke": Variable(
        "meso-scale energy",
        W_GRID + TIMESTEPS,
        "m^2/s^2",
        "meso-scale energy",
        write_to_restart=True,
        active=lambda settings: settings.enable_eke,
    ),
    "deke": Variable(
        "meso-scale energy tendency",
        W_GRID + TIMESTEPS,
        "m^2/s^3",
        "meso-scale energy tendency",
        write_to_restart=True,
        active=lambda settings: settings.enable_eke,
    ),
    "sqrteke": Variable(
        "square-root of eke",
        W_GRID,
        "m/s",
        "square-root of eke",
        active=lambda settings: settings.enable_eke,
    ),
    "L_rossby": Variable(
        "Rossby radius",
        T_HOR,
        "m",
        "Rossby radius",
        active=lambda settings: settings.enable_eke,
    ),
    "L_rhines": Variable(
        "Rhines scale",
        W_GRID,
        "m",
        "Rhines scale",
        active=lambda settings: settings.enable_eke,
    ),
    "eke_len": Variable(
        "Eddy length scale",
        T_GRID,
        "m",
        "Eddy length scale",
        active=lambda settings: settings.enable_eke,
    ),
    "eke_diss_iw": Variable(
        "Dissipation of EKE to IW",
        W_GRID,
        "m^2/s^3",
        "Dissipation of EKE to internal waves",
        active=lambda settings: settings.enable_eke,
    ),
    "eke_diss_tke": Variable(
        "Dissipation of EKE to TKE",
        W_GRID,
        "m^2/s^3",
        "Dissipation of EKE to TKE",
        active=lambda settings: settings.enable_eke,
    ),
    "E_iw": Variable(
        "Internal wave energy",
        W_GRID + TIMESTEPS,
        "m^2/s^2",
        "Internal wave energy",
        write_to_restart=True,
        active=lambda settings: settings.enable_idemix,
    ),
    "dE_iw": Variable(
        "Internal wave energy tendency",
        W_GRID + TIMESTEPS,
        "m^2/s^2",
        "Internal wave energy tendency",
        write_to_restart=True,
        active=lambda settings: settings.enable_idemix,
    ),
    "c0": Variable(
        "Vertical IW group velocity",
        W_GRID,
        "m/s",
        "Vertical internal wave group velocity",
        active=lambda settings: settings.enable_idemix,
    ),
    "v0": Variable(
        "Horizontal IW group velocity",
        W_GRID,
        "m/s",
        "Horizontal internal wave group velocity",
        active=lambda settings: settings.enable_idemix,
    ),
    "alpha_c": Variable(
        "?",
        W_GRID,
        "?",
        "?",
        active=lambda settings: settings.enable_idemix,
    ),
    "iw_diss": Variable(
        "IW dissipation",
        W_GRID,
        "m^2/s^3",
        "Internal wave dissipation",
        active=lambda settings: settings.enable_idemix,
    ),
    "forc_iw_surface": Variable(
        "IW surface forcing",
        T_HOR,
        "m^3/s^3",
        "Internal wave surface forcing",
        time_dependent=False,
        active=lambda settings: settings.enable_idemix,
    ),
    "forc_iw_bottom": Variable(
        "IW bottom forcing",
        T_HOR,
        "m^3/s^3",
        "Internal wave bottom forcing",
        time_dependent=False,
        active=lambda settings: settings.enable_idemix,
    ),
}


def manifest_metadata(var_meta, settings):
    """Evaluate callable metadata fields given the current settings."""
    from copy import copy

    out = {}

    for var_name, var_val in var_meta.items():
        var_val = copy(var_val)
        for attr, attr_val in vars(var_val).items():
            if callable(attr_val) and attr != "get_mask":
                setattr(var_val, attr, attr_val(settings))

        out[var_name] = var_val

    return out


def allocate(dimensions, grid, dtype=None, include_ghosts=True, local=True, fill=0):
    from veros.core.operators import numpy as npx

    if dtype is None:
        dtype = runtime_settings.float_type

    shape = get_shape(dimensions, grid, include_ghosts=include_ghosts, local=local)
    out = npx.full(shape, fill, dtype=dtype)

    if runtime_settings.backend == "numpy":
        out.flags.writeable = False

    return out



from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.core import advection, utilities
from veros.core.operators import update, update_add, at, for_loop, numpy as npx


@veros_routine
def set_tke_diffusivities(state):
    vs = state.variables
    settings = state.settings

    if settings.enable_tke:
        tke_diff_out = set_tke_diffusivities_kernel(state)
        vs.update(tke_diff_out)
    else:
        vs.kappaM = update(vs.kappaM, at[...], settings.kappaM_0)
        vs.kappaH = npx.where(vs.Nsqr[..., vs.tau] < 0.0, 1.0, settings.kappaH_0)


@veros_kernel
def set_tke_diffusivities_kernel(state):
    """
    set vertical diffusivities based on TKE model
    """
    vs = state.variables
    settings = state.settings

    Rinumber = allocate(state.dimensions, ("xt", "yt", "zt"))

    vs.sqrttke = npx.sqrt(npx.maximum(0.0, vs.tke[:, :, :, vs.tau]))
    """
    calculate buoyancy length scale
    """
    vs.mxl = npx.sqrt(2) * vs.sqrttke / npx.sqrt(npx.maximum(1e-12, vs.Nsqr[:, :, :, vs.tau])) * vs.maskW

    """
    apply limits for mixing length
    """
    if settings.tke_mxl_choice == 1:
        """
        bounded by the distance to surface/bottom
        """
        vs.mxl = npx.minimum(
            npx.minimum(vs.mxl, -vs.zw[npx.newaxis, npx.newaxis, :] + vs.dzw[npx.newaxis, npx.newaxis, :] * 0.5),
            vs.ht[:, :, npx.newaxis] + vs.zw[npx.newaxis, npx.newaxis, :],
        )
        vs.mxl = npx.maximum(vs.mxl, settings.mxl_min)
    elif settings.tke_mxl_choice == 2:
        """
        bound length scale as in mitgcm/OPA code
        """
        nz = state.dimensions["zt"]

        def backwards_pass(kinv, mxl):
            k = nz - kinv - 1
            return update(mxl, at[:, :, k], npx.minimum(mxl[:, :, k], mxl[:, :, k + 1] + vs.dzt[k + 1]))

        vs.mxl = for_loop(1, nz, backwards_pass, vs.mxl)
        vs.mxl = update(vs.mxl, at[:, :, -1], npx.minimum(vs.mxl[:, :, -1], settings.mxl_min + vs.dzt[-1]))

        def forwards_pass(k, mxl):
            return update(mxl, at[:, :, k], npx.minimum(mxl[:, :, k], mxl[:, :, k - 1] + vs.dzt[k]))

        vs.mxl = for_loop(1, nz, forwards_pass, vs.mxl)
        vs.mxl = npx.maximum(vs.mxl, settings.mxl_min)
    else:
        raise ValueError("unknown mixing length choice in tke_mxl_choice")

    """
    calculate viscosity and diffusivity based on Prandtl number
    """
    vs.K_diss_v = utilities.enforce_boundaries(vs.K_diss_v, settings.enable_cyclic_x)
    vs.kappaM = update(vs.kappaM, at[...], npx.minimum(settings.kappaM_max, settings.c_k * vs.mxl * vs.sqrttke))
    Rinumber = update(
        Rinumber, at[...], vs.Nsqr[:, :, :, vs.tau] / npx.maximum(vs.K_diss_v / npx.maximum(1e-12, vs.kappaM), 1e-12)
    )
    if settings.enable_idemix:
        Rinumber = update(
            Rinumber,
            at[...],
            npx.minimum(
                Rinumber,
                vs.kappaM * vs.Nsqr[:, :, :, vs.tau] / npx.maximum(1e-12, vs.alpha_c * vs.E_iw[:, :, :, vs.tau] ** 2),
            ),
        )

    if settings.enable_Prandtl_tke:
        vs.Prandtlnumber = npx.maximum(1.0, npx.minimum(10, 6.6 * Rinumber))
    else:
        vs.Prandtlnumber = update(vs.Prandtlnumber, at[...], settings.Prandtl_tke0)

    vs.kappaH = npx.maximum(settings.kappaH_min, vs.kappaM / vs.Prandtlnumber)

    if settings.enable_kappaH_profile:
        # Correct diffusivity according to
        # Bryan, K., and L. J. Lewis, 1979:
        # A water mass model of the world ocean. J. Geophys. Res., 84, 2503–2517.
        # It mainly modifies kappaH within 20S - 20N deg. belt
        vs.kappaH = npx.maximum(
            vs.kappaH,
            (0.8 + 1.05 / settings.pi * npx.arctan((-vs.zw[npx.newaxis, npx.newaxis, :] - 2500.0) / 222.2)) * 1e-4,
        )

    vs.kappaM = npx.maximum(settings.kappaM_min, vs.kappaM)

    return KernelOutput(
        sqrttke=vs.sqrttke,
        mxl=vs.mxl,
        kappaM=vs.kappaM,
        kappaH=vs.kappaH,
        Prandtlnumber=vs.Prandtlnumber,
        K_diss_v=vs.K_diss_v,
    )


@veros_routine
def integrate_tke(state):
    vs = state.variables
    tke_out = integrate_tke_kernel(state)
    vs.update(tke_out)


@veros_kernel
def integrate_tke_kernel(state):
    """
    integrate Tke equation on W grid with surface flux boundary condition
    """
    vs = state.variables
    settings = state.settings

    conditional_outputs = {}

    flux_east = allocate(state.dimensions, ("xt", "yt", "zt"))
    flux_north = allocate(state.dimensions, ("xt", "yt", "zt"))
    flux_top = allocate(state.dimensions, ("xt", "yt", "zt"))

    dt_tke = settings.dt_mom  # use momentum time step to prevent spurious oscillations

    """
    Sources and sinks by vertical friction, vertical mixing, and non-conservative advection
    """
    forc = vs.K_diss_v - vs.P_diss_v - vs.P_diss_adv

    """
    store transfer due to vertical mixing from dyn. enthalpy by non-linear eq.of
    state either to TKE or to heat
    """
    if not settings.enable_store_cabbeling_heat:
        forc = forc - vs.P_diss_nonlin

    """
    transfer part of dissipation of EKE to TKE
    """
    if settings.enable_eke:
        forc = forc + vs.eke_diss_tke

    if settings.enable_idemix:
        """
        transfer dissipation of internal waves to TKE
        """
        forc = forc + vs.iw_diss
        """
        store bottom friction either in TKE or internal waves
        """
        if settings.enable_store_bottom_friction_tke:
            forc = forc + vs.K_diss_bot

    else:  # short-cut without idemix
        if settings.enable_eke:
            forc = forc + vs.eke_diss_iw

        else:  # and without EKE model
            if settings.enable_store_cabbeling_heat:
                forc = forc + vs.K_diss_gm + vs.K_diss_h - vs.P_diss_skew - vs.P_diss_hmix - vs.P_diss_iso
            else:
                forc = forc + vs.K_diss_gm + vs.K_diss_h - vs.P_diss_skew

        forc = forc + vs.K_diss_bot

    """
    vertical mixing and dissipation of TKE
    """
    _, water_mask, edge_mask = utilities.create_water_masks(vs.kbot[2:-2, 2:-2], settings.nz)

    a_tri, b_tri, c_tri, d_tri, delta = (
        allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2, :] for _ in range(5)
    )

    delta = update(
        delta,
        at[:, :, :-1],
        dt_tke
        / vs.dzt[npx.newaxis, npx.newaxis, 1:]
        * settings.alpha_tke
        * 0.5
        * (vs.kappaM[2:-2, 2:-2, :-1] + vs.kappaM[2:-2, 2:-2, 1:]),
    )

    a_tri = update(a_tri, at[:, :, 1:-1], -delta[:, :, :-2] / vs.dzw[npx.newaxis, npx.newaxis, 1:-1])
    a_tri = update(a_tri, at[:, :, -1], -delta[:, :, -2] / (0.5 * vs.dzw[-1]))

    b_tri = update(
        b_tri,
        at[:, :, 1:-1],
        1
        + (delta[:, :, 1:-1] + delta[:, :, :-2]) / vs.dzw[npx.newaxis, npx.newaxis, 1:-1]
        + dt_tke * settings.c_eps * vs.sqrttke[2:-2, 2:-2, 1:-1] / vs.mxl[2:-2, 2:-2, 1:-1],
    )
    b_tri = update(
        b_tri,
        at[:, :, -1],
        1
        + delta[:, :, -2] / (0.5 * vs.dzw[-1])
        + dt_tke * settings.c_eps / vs.mxl[2:-2, 2:-2, -1] * vs.sqrttke[2:-2, 2:-2, -1],
    )
    b_tri_edge = (
        1
        + delta / vs.dzw[npx.newaxis, npx.newaxis, :]
        + dt_tke * settings.c_eps / vs.mxl[2:-2, 2:-2, :] * vs.sqrttke[2:-2, 2:-2, :]
    )

    c_tri = update(c_tri, at[:, :, :-1], -delta[:, :, :-1] / vs.dzw[npx.newaxis, npx.newaxis, :-1])

    d_tri = update(d_tri, at[...], vs.tke[2:-2, 2:-2, :, vs.tau] + dt_tke * forc[2:-2, 2:-2, :])
    d_tri = update_add(d_tri, at[:, :, -1], dt_tke * vs.forc_tke_surface[2:-2, 2:-2] / (0.5 * vs.dzw[-1]))

    sol = utilities.solve_implicit(a_tri, b_tri, c_tri, d_tri, water_mask, b_edge=b_tri_edge, edge_mask=edge_mask)
    vs.tke = update(vs.tke, at[2:-2, 2:-2, :, vs.taup1], npx.where(water_mask, sol, vs.tke[2:-2, 2:-2, :, vs.taup1]))

    """
    store tke dissipation for diagnostics
    """
    vs.tke_diss = settings.c_eps / vs.mxl * vs.sqrttke * vs.tke[:, :, :, vs.taup1]

    """
    Add TKE if surface density flux drains TKE in uppermost box
    """
    mask = vs.tke[2:-2, 2:-2, -1, vs.taup1] < 0.0
    vs.tke_surf_corr = update(
        vs.tke_surf_corr,
        at[2:-2, 2:-2],
        npx.where(mask, -vs.tke[2:-2, 2:-2, -1, vs.taup1] * 0.5 * vs.dzw[-1] / dt_tke, 0.0),
    )
    vs.tke = update(vs.tke, at[2:-2, 2:-2, -1, vs.taup1], npx.maximum(0.0, vs.tke[2:-2, 2:-2, -1, vs.taup1]))

    if settings.enable_tke_hor_diffusion:
        """
        add tendency due to lateral diffusion
        """
        flux_east = update(
            flux_east,
            at[:-1, :, :],
            settings.K_h_tke
            * (vs.tke[1:, :, :, vs.tau] - vs.tke[:-1, :, :, vs.tau])
            / (vs.cost[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
            * vs.maskU[:-1, :, :],
        )

        flux_north = update(
            flux_north,
            at[:, :-1, :],
            settings.K_h_tke
            * (vs.tke[:, 1:, :, vs.tau] - vs.tke[:, :-1, :, vs.tau])
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * vs.maskV[:, :-1, :]
            * vs.cosu[npx.newaxis, :-1, npx.newaxis],
        )
        flux_north = update(flux_north, at[:, -1, :], 0.0)

        vs.tke = update_add(
            vs.tke,
            at[2:-2, 2:-2, :, vs.taup1],
            dt_tke
            * vs.maskW[2:-2, 2:-2, :]
            * (
                (flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
                + (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
            ),
        )

    """
    add tendency due to advection
    """
    if settings.enable_tke_superbee_advection:
        flux_east, flux_north, flux_top = advection.adv_flux_superbee_wgrid(state, vs.tke[:, :, :, vs.tau])

    if settings.enable_tke_upwind_advection:
        flux_east, flux_north, flux_top = advection.adv_flux_upwind_wgrid(state, vs.tke[:, :, :, vs.tau])

    if settings.enable_tke_superbee_advection or settings.enable_tke_upwind_advection:
        vs.dtke = update(
            vs.dtke,
            at[2:-2, 2:-2, :, vs.tau],
            vs.maskW[2:-2, 2:-2, :]
            * (
                -(flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
                - (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
            ),
        )
        vs.dtke = update_add(vs.dtke, at[:, :, 0, vs.tau], -flux_top[:, :, 0] / vs.dzw[0])
        vs.dtke = update_add(
            vs.dtke, at[:, :, 1:-1, vs.tau], -(flux_top[:, :, 1:-1] - flux_top[:, :, :-2]) / vs.dzw[1:-1]
        )
        vs.dtke = update_add(
            vs.dtke, at[:, :, -1, vs.tau], -(flux_top[:, :, -1] - flux_top[:, :, -2]) / (0.5 * vs.dzw[-1])
        )

        """
        Adam Bashforth time stepping
        """
        vs.tke = update_add(
            vs.tke,
            at[:, :, :, vs.taup1],
            settings.dt_tracer
            * (
                (1.5 + settings.AB_eps) * vs.dtke[:, :, :, vs.tau]
                - (0.5 + settings.AB_eps) * vs.dtke[:, :, :, vs.taum1]
            ),
        )

        conditional_outputs.update(dtke=vs.dtke)

    return KernelOutput(tke=vs.tke, tke_surf_corr=vs.tke_surf_corr, tke_diss=vs.tke_diss, **conditional_outputs)



"""
==========================================================================
  linear equation of state
  input is Salinity sa in g/kg,
  pot. temperature ct in deg C
==========================================================================
"""
from veros import veros_kernel

rho0 = 1024.0
theta0 = 283.0 - 273.15
S0 = 35.0
betaT = 1.67e-4
betaS = 0.78e-3
grav = 9.81
z0 = 0.0


@veros_kernel
def linear_eq_of_state_rho(sa, ct):
    return -(betaT * (ct - theta0) - betaS * (sa - S0)) * rho0


@veros_kernel
def linear_eq_of_state_dyn_enthalpy(sa, ct, p):
    zz = -p - z0
    thetas = ct - theta0
    return grav * zz * (-betaT * thetas + betaS * (sa - S0))


@veros_kernel
def linear_eq_of_state_salt(rho, ct):
    return (rho + betaT * (ct - theta0) * rho0) / (betaS * rho0) + S0


@veros_kernel
def linear_eq_of_state_drhodT():
    return -betaT * rho0


@veros_kernel
def linear_eq_of_state_drhodS():
    return betaS * rho0


@veros_kernel
def linear_eq_of_state_drhodp():
    return 0.0



from veros.core.operators import numpy as npx
import numpy as onp

import scipy.interpolate
import scipy.spatial


def interpolate(coords, var, interp_coords, missing_value=None, fill=True, kind="linear"):
    """Interpolate globally defined data to a different (regular) grid.

    Arguments:
       coords: Tuple of coordinate arrays for each dimension.
       var (:obj:`ndarray` of dim (nx1, ..., nxd)): Variable data to interpolate.
       interp_coords: Tuple of coordinate arrays to interpolate to.
       missing_value (optional): Value denoting cells of missing data in ``var``.
          Is replaced by `NaN` before interpolating. Defaults to `None`, which means
          no replacement is taking place.
       fill (bool, optional): Whether `NaN` values should be replaced by the nearest
          finite value after interpolating. Defaults to ``True``.
       kind (str, optional): Order of interpolation. Supported are `nearest` and
          `linear` (default).

    Returns:
       :obj:`ndarray` containing the interpolated values on the grid spanned by
       ``interp_coords``.

    """
    if len(coords) != len(interp_coords) or len(coords) != var.ndim:
        raise ValueError("Dimensions of coordinates and values do not match")

    if missing_value is not None:
        invalid_mask = npx.isclose(var, missing_value)
        var = npx.where(invalid_mask, npx.nan, var)

    if var.ndim > 1 and coords[0].ndim == 1:
        interp_grid = npx.rollaxis(npx.array(npx.meshgrid(*interp_coords, indexing="ij")), 0, len(interp_coords) + 1)
    else:
        interp_grid = interp_coords

    coords = [onp.array(c) for c in coords]
    var = scipy.interpolate.interpn(
        coords, onp.array(var), interp_grid, bounds_error=False, fill_value=npx.nan, method=kind
    )
    var = npx.asarray(var)

    if fill:
        var = fill_holes(var)

    return var


def fill_holes(data):
    """A simple inpainting function that replaces NaN values in `data` with the
    nearest finite value.
    """
    data = onp.array(data)
    dim = data.ndim
    flag = ~onp.isnan(data)

    slcs = [slice(None)] * dim

    while onp.any(~flag):
        for i in range(dim):
            slcs1 = slcs[:]
            slcs2 = slcs[:]
            slcs1[i] = slice(0, -1)
            slcs2[i] = slice(1, None)

            slcs1 = tuple(slcs1)
            slcs2 = tuple(slcs2)

            # replace from the right
            repmask = onp.logical_and(~flag[slcs1], flag[slcs2])
            data[slcs1][repmask] = data[slcs2][repmask]
            flag[slcs1][repmask] = True

            # replace from the left
            repmask = onp.logical_and(~flag[slcs2], flag[slcs1])
            data[slcs2][repmask] = data[slcs1][repmask]
            flag[slcs2][repmask] = True

    return npx.asarray(data)


def get_periodic_interval(current_time, cycle_length, rec_spacing, n_rec):
    """Used for linear interpolation between periodic time intervals.

    One common application is the interpolation of external forcings that are defined
    at discrete times (e.g. one value per month of a standard year) to the current
    time step.

    Arguments:
       current_time (float): Time to interpolate to.
       cycle_length (float): Total length of one periodic cycle.
       rec_spacing (float): Time spacing between each data record.
       n_rec (int): Total number of records available.

    Returns:
       :obj:`tuple` containing (n1, f1), (n2, f2): Indices and weights for the interpolated
       record array.

    Example:
       The following interpolates a record array ``data`` containing 12 monthly values
       to the current time step:

       >>> year_in_seconds = 60. * 60. * 24. * 365.
       >>> current_time = 60. * 60. * 24. * 45. # mid-february
       >>> print(data.shape)
       (360, 180, 12)
       >>> (n1, f1), (n2, f2) = get_periodic_interval(current_time, year_in_seconds, year_in_seconds / 12, 12)
       >>> data_at_current_time = f1 * data[..., n1] + f2 * data[..., n2]

    """
    current_time = current_time % cycle_length
    # using npx.array works with both NumPy and JAX
    t_idx_1 = npx.array(current_time // rec_spacing, dtype="int")
    t_idx_2 = npx.array((1 + t_idx_1) % n_rec, dtype="int")
    weight_2 = (current_time - rec_spacing * t_idx_1) / rec_spacing
    weight_1 = 1.0 - weight_2
    return (t_idx_1, weight_1), (t_idx_2, weight_2)


def make_cyclic(longitude, array=None, wrap=360.0):
    """Create a cyclic version of a longitude array and (optionally) another array.

    Arguments:
        longitude (ndarray): Longitude array of shape (nlon, ...).
        array (ndarray): Another array that is to be made cyclic of shape (nlon, ...).
        wrap (float): Wrapping value, defaults to 360 (degrees).

    Returns:
        Tuple containing (cyclic_longitudes, cyclic_array) if `array` is given, otherwise
        just the ndarray cyclic_longitudes of shape (2 * nlon, ...).

    """
    lonsize = longitude.shape[0]
    cyclic_longitudes = npx.hstack(
        (longitude[lonsize // 2 :, ...] - wrap, longitude, longitude[: lonsize // 2, ...] + wrap)
    )
    if array is None:
        return cyclic_longitudes
    cyclic_array = npx.hstack((array[lonsize // 2 :, ...], array, array[: lonsize // 2, ...]))
    return cyclic_longitudes, cyclic_array


def get_coastline_distance(coords, coast_mask, spherical=False, radius=None, num_candidates=None):
    """Calculate the (approximate) distance of each water cell from the nearest coastline.

    Arguments:
        coords (tuple of ndarrays): Tuple containing x and y (longitude and latitude)
            coordinate arrays of shape (nx, ny).
        coast_mask (ndarray): Boolean mask indicating whether a cell is a land cell
            (must be same shape as coordinate arrays).
        spherical (bool): Use spherical instead of Cartesian coordinates.
            When this is `True`, cyclical boundary conditions are used, and the
            resulting distances are only approximate. Cells are pre-sorted by
            Euclidean lon-lat distance, and great circle distances are calculated for
            the first `num_candidates` elements. Defaults to `False`.
        radius (float): Radius of spherical coordinate system. Must be given when
            `spherical` is `True`.
        num_candidates (int): Number of candidates to calculate great circle distances
            for for each water cell. The higher this value, the more accurate the returned
            distances become when `spherical` is `True`. Defaults to the square root
            of the number of coastal cells.

    Returns:
        :obj:`ndarray` of shape (nx, ny) indicating the distance to the nearest land
        cell (0 if cell is land).

    Example:
        The following returns coastal distances of all T cells for a spherical Veros setup.

        >>> coords = npx.meshgrid(vs.xt[2:-2], vs.yt[2:-2], indexing='ij')
        >>> dist = tools.get_coastline_distance(coords, vs.kbot > 0, spherical=True, radius=settings.radius)

    """
    if not len(coords) == 2:
        raise ValueError("coords must be lon-lat tuple")
    if not all(c.shape == coast_mask.shape for c in coords):
        raise ValueError("coordinates must have same shape as coastal mask")
    if spherical and not radius:
        raise ValueError("radius must be given for spherical coordinates")

    watercoords = onp.array([c[~coast_mask] for c in coords]).T
    if spherical:
        coastcoords = onp.array(make_cyclic(coords[0][coast_mask], coords[1][coast_mask])).T
    else:
        coastcoords = onp.array((coords[0][coast_mask], coords[1][coast_mask])).T
    coast_kdtree = scipy.spatial.cKDTree(coastcoords)

    distance = onp.zeros(coords[0].shape)

    if spherical:

        def spherical_distance(coords1, coords2):
            """Calculate great circle distance from latitude and longitude"""
            coords1 *= onp.pi / 180.0
            coords2 *= onp.pi / 180.0
            lon1, lon2, lat1, lat2 = coords1[..., 0], coords2[..., 0], coords1[..., 1], coords2[..., 1]
            return radius * onp.arccos(
                onp.sin(lat1) * onp.sin(lat2) + onp.cos(lat1) * onp.cos(lat2) * onp.cos(lon1 - lon2)
            )

        if not num_candidates:
            num_candidates = int(onp.sqrt(onp.count_nonzero(~coast_mask)))

        i_nearest = coast_kdtree.query(watercoords, k=num_candidates)[1]
        approx_nearest = coastcoords[i_nearest]
        distance[~coast_mask] = onp.min(spherical_distance(approx_nearest, watercoords[..., onp.newaxis, :]), axis=-1)

    else:
        distance[~coast_mask] = coast_kdtree.query(watercoords)[0]

    return npx.asarray(distance)


def get_uniform_grid_steps(total_length, stepsize):
    """Get uniform grid step sizes in an interval.

    Arguments:
        total_length (float): total length of the resulting grid
        stepsize (float): grid step size

    Returns:
        :obj:`ndarray` of grid steps

    Example:
        >>> uniform_steps = uniform_grid_setup(6., 0.25)
        >>> uniform_steps
        [ 0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,
          0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.25,
          0.25,  0.25,  0.25,  0.25,  0.25,  0.25 ]

    """
    if total_length % stepsize:
        raise ValueError("total length must be an integer multiple of stepsize")
    return stepsize * npx.ones(int(total_length / stepsize))


def get_stretched_grid_steps(
    n_cells, total_length, minimum_stepsize, stretching_factor=2.5, two_sided_grid=False, refine_towards="upper"
):
    """Computes stretched grid steps for regional and global domains with either
    one or two-sided stretching using a hyperbolic tangent stretching function.

    Arguments:
        n_cells (int): Number of grid points.
        total_length (float): Length of the grid interval to be covered (sum of the
            resulting grid steps).
        minimum_stepsize (float): Grid step size on the lower end of the interval.
        stretching_factor (float, optional): Coefficient of the `tanh` stretching
            function. The higher this value, the more abrupt the step sizes change.
        two_sided_grid (bool, optional): If set to `True`, the resulting grid will be symmetrical
            around the center. Defaults to `False`.
        refine_towards ('upper' or 'lower', optional): The side of the interval that is to be refined.
            Defaults to 'upper'.

    Returns:
        :obj:`ndarray` of shape `(n_cells)` containing grid steps.

    Examples:
        >>> dyt = get_stretched_grid_steps(14, 180, 5)
        >>> dyt
        [  5.10517337   5.22522948   5.47813251   5.99673813   7.00386752
           8.76808565  11.36450896  14.34977676  16.94620006  18.71041819
          19.71754758  20.2361532   20.48905624  20.60911234]
        >>> dyt.sum()
        180.0

        >>> dyt = get_stretched_grid_steps(14, 180, 5, stretching_factor=4.)
        >>> dyt
        [  5.00526979   5.01802837   5.06155549   5.20877528   5.69251688
           7.14225176  10.51307232  15.20121339  18.57203395  20.02176884
          20.50551044  20.65273022  20.69625734  20.70901593]
        >>> dyt.sum()
        180.0

    """

    if refine_towards not in ("upper", "lower"):
        raise ValueError('refine_towards must be "upper" or "lower"')
    if two_sided_grid:
        if n_cells % 2:
            raise ValueError(f"number of grid points must be even integer number (given: {n_cells})")
        n_cells = n_cells / 2

    stretching_function = npx.tanh(stretching_factor * npx.linspace(-1, 1, n_cells))

    if refine_towards == "lower":
        stretching_function = stretching_function[::-1]
    if two_sided_grid:
        stretching_function = npx.concatenate((stretching_function[::-1], stretching_function))

    def normalize_sum(var, sum_value, minimum_value=0.0):
        if abs(var.sum()) < 1e-5:
            var += 1
        var *= (sum_value - len(var) * minimum_value) / var.sum()
        return var + minimum_value

    stretching_function = normalize_sum(stretching_function, total_length, minimum_stepsize)
    assert abs(1 - npx.sum(stretching_function) / total_length) < 1e-5, "precision error"
    return stretching_function


def get_vinokur_grid_steps(
    n_cells, total_length, lower_stepsize, upper_stepsize=None, two_sided_grid=False, refine_towards="upper"
):
    """Computes stretched grid steps for regional and global domains with either
    one or two-sided stretching using Vinokur stretching.

    This stretching function minimizes discretization errors on finite difference
    grids.

    Arguments:
        n_cells (int): Number of grid points.
        total_length (float): Length of the grid interval to be covered (sum of the
            resulting grid steps).
        lower_stepsize (float): Grid step size on the lower end of the interval.
        upper_stepsize (float or ``None``, optional): Grid step size on the upper end of the interval.
            If not given, the one-sided version of the algorithm is used (that enforces zero curvature
            on the upper end).
        two_sided_grid (bool, optional): If set to `True`, the resulting grid will be symmetrical
            around the center. Defaults to `False`.
        refine_towards ('upper' or 'lower', optional): The side of the interval that is to be refined.
            Defaults to 'upper'.

    Returns:
        :obj:`ndarray` of shape `(n_cells)` containing grid steps.

    Reference:
        Vinokur, Marcel, On One-Dimensional Stretching Functions for Finite-Difference Calculations,
        Journal of Computational Physics. 50, 215, 1983.

    Examples:
        >>> dyt = get_vinokur_grid_steps(14, 180, 5, two_sided_grid=True)
        >>> dyt
        [ 18.2451554   17.23915939  15.43744632  13.17358802  10.78720589
           8.53852027   6.57892471   6.57892471   8.53852027  10.78720589
          13.17358802  15.43744632  17.23915939  18.2451554 ]
        >>> dyt.sum()
        180.

        >>> dyt = get_vinokur_grid_steps(14, 180, 5, upper_stepsize=10)
        >>> dyt
        [  5.9818365    7.3645667    8.92544833  10.61326984  12.33841985
          13.97292695  15.36197306  16.3485688   16.80714121  16.67536919
          15.97141714  14.78881918  13.27136448  11.57887877 ]
        >>> dyt.sum()
        180.

    """
    if refine_towards not in ("upper", "lower"):
        raise ValueError('refine_towards must be "upper" or "lower"')
    if two_sided_grid:
        if n_cells % 2:
            raise ValueError(f"number of grid points must be an even integer (given: {n_cells})")
        n_cells = n_cells // 2

    n_cells += 1

    def approximate_sinc_inverse(y):
        """Approximate inverse of sin(y) / y"""
        if y < 0.26938972:
            inv = npx.pi * (
                1
                - y
                + y**2
                - (1 + npx.pi**2 / 6) * y**3
                + 6.794732 * y**4
                - 13.205501 * y**5
                + 11.726095 * y**6
            )
        else:
            ybar = 1.0 - y
            inv = npx.sqrt(6 * ybar) * (
                1
                + 0.15 * ybar
                + 0.057321429 * ybar**2
                + 0.048774238 * ybar**3
                - 0.053337753 * ybar**4
                + 0.075845134 * ybar**5
            )
        assert abs(1 - npx.sin(inv) / inv / y) < 1e-2, "precision error"
        return inv

    def approximate_sinhc_inverse(y):
        """Approximate inverse of sinh(y) / y"""
        if y < 2.7829681:
            ybar = y - 1.0
            inv = npx.sqrt(6 * ybar) * (
                1
                - 0.15 * ybar
                + 0.057321429 * ybar**2
                - 0.024907295 * ybar**3
                + 0.0077424461 * ybar**4
                - 0.0010794123 * ybar**5
            )
        else:
            v = npx.log(y)
            w = 1.0 / y - 0.028527431
            inv = (
                v
                + (1 + 1.0 / v) * npx.log(2 * v)
                - 0.02041793
                + 0.24902722 * w
                + 1.9496443 * w**2
                - 2.6294547 * w**3
                + 8.56795911 * w**4
            )
        assert abs(1 - npx.sinh(inv) / inv / y) < 1e-2, "precision error"
        return inv

    target_sum = total_length
    if two_sided_grid:
        target_sum *= 0.5

    s0 = float(target_sum) / float(lower_stepsize * n_cells)
    if upper_stepsize:
        s1 = float(target_sum) / float(upper_stepsize * n_cells)
        a, b = npx.sqrt(s1 / s0), npx.sqrt(s1 * s0)
        if b > 1:
            stretching_factor = approximate_sinhc_inverse(b)
            stretched_grid = 0.5 + 0.5 * npx.tanh(stretching_factor * npx.linspace(-0.5, 0.5, n_cells)) / npx.tanh(
                0.5 * stretching_factor
            )
        else:
            stretching_factor = approximate_sinc_inverse(b)
            stretched_grid = 0.5 + 0.5 * npx.tan(stretching_factor * npx.linspace(-0.5, 0.5, n_cells)) / npx.tan(
                0.5 * stretching_factor
            )
        stretched_grid = stretched_grid / (a + (1.0 - a) * stretched_grid)
    else:
        if s0 > 1:
            stretching_factor = approximate_sinhc_inverse(s0) * 0.5
            stretched_grid = 1 + npx.tanh(stretching_factor * npx.linspace(0.0, 1.0, n_cells)) / npx.tanh(
                stretching_factor
            )
        else:
            stretching_factor = approximate_sinc_inverse(s0) * 0.5
            stretched_grid = 1 + npx.tan(stretching_factor * npx.linspace(0.0, 1.0, n_cells)) / npx.tan(
                stretching_factor
            )

    stretched_grid_steps = npx.diff(stretched_grid * target_sum)
    if refine_towards == "upper":
        stretched_grid_steps = stretched_grid_steps[::-1]
    if two_sided_grid:
        stretched_grid_steps = npx.concatenate((stretched_grid_steps[::-1], stretched_grid_steps))

    assert abs(1 - npx.sum(stretched_grid_steps) / total_length) < 1e-5, "precision error"
    return stretched_grid_steps



"""Veros, the versatile ocean simulator"""

import sys
import types

# black magic: ensure lazy imports for public API by overriding module.__class__


def _reraise_exceptions(func):
    import functools

    @functools.wraps(func)
    def reraise_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise ImportError("Critical error during initial import") from e

    return reraise_wrapper


class _PublicAPI(types.ModuleType):
    @property
    @_reraise_exceptions
    def __version__(self):
        from veros._version import get_versions

        return get_versions()["version"]

    @property
    @_reraise_exceptions
    def logger(self):
        if not hasattr(self, "_logger"):
            from veros.logs import setup_logging

            self._logger = setup_logging()
        return self._logger

    @property
    @_reraise_exceptions
    def runtime_settings(self):
        if not hasattr(self, "_runtime_settings"):
            from veros.runtime import RuntimeSettings

            self._runtime_settings = RuntimeSettings()
        return self._runtime_settings

    @property
    @_reraise_exceptions
    def runtime_state(self):
        if not hasattr(self, "_runtime_state"):
            from veros.runtime import RuntimeState

            self._runtime_state = RuntimeState()
        return self._runtime_state

    @property
    @_reraise_exceptions
    def veros_routine(self):
        from veros.routines import veros_routine

        return veros_routine

    @property
    @_reraise_exceptions
    def veros_kernel(self):
        from veros.routines import veros_kernel

        return veros_kernel

    @property
    @_reraise_exceptions
    def KernelOutput(self):
        from veros.state import KernelOutput

        return KernelOutput

    @property
    @_reraise_exceptions
    def VerosSetup(self):
        from veros.veros import VerosSetup

        return VerosSetup

    @property
    @_reraise_exceptions
    def VerosState(self):
        from veros.state import VerosState

        return VerosState


sys.modules[__name__].__class__ = _PublicAPI

del sys
del types
del _PublicAPI
del _reraise_exceptions



import warnings

BACKENDS = ("numpy", "jax")

BACKEND_MESSAGES = {"jax": "Kernels are compiled during first iteration, be patient"}

_init_done = set()


def init_jax_config():
    if "jax" in _init_done:
        return

    import jax
    from veros import runtime_settings, runtime_state
    from veros.state import (
        VerosState,
        VerosVariables,
        DistSafeVariableWrapper,
        veros_state_pytree_flatten,
        veros_state_pytree_unflatten,
        veros_variables_pytree_flatten,
        veros_variables_pytree_unflatten,
        dist_safe_wrapper_pytree_flatten,
        dist_safe_wrapper_pytree_unflatten,
    )

    if runtime_state.proc_num > 1:
        try:
            import mpi4jax  # noqa: F401
        except ImportError as exc:
            raise RuntimeError("Running JAX with MPI requires mpi4jax to be installed") from exc

    if runtime_settings.float_type == "float64":
        jax.config.update("jax_enable_x64", True)
    else:
        # ignore warnings about unavailable x64 types
        warnings.filterwarnings("ignore", message="Explicitly requested dtype.*", module="jax")

    jax.config.update("jax_platform_name", runtime_settings.device)

    jax.tree_util.register_pytree_node(VerosState, veros_state_pytree_flatten, veros_state_pytree_unflatten)
    jax.tree_util.register_pytree_node(VerosVariables, veros_variables_pytree_flatten, veros_variables_pytree_unflatten)
    jax.tree_util.register_pytree_node(
        DistSafeVariableWrapper, dist_safe_wrapper_pytree_flatten, dist_safe_wrapper_pytree_unflatten
    )

    _init_done.add("jax")


def get_backend_module(backend_name):
    if backend_name not in BACKENDS:
        raise ValueError(f"unrecognized backend {backend_name} (must be either of: {list(BACKENDS.keys())!r})")

    if backend_name == "jax":
        init_jax_config()
        import jax.numpy as backend_module

    elif backend_name == "numpy":
        import numpy as backend_module

    return backend_module


def get_curent_device_name():
    from veros import runtime_settings

    if runtime_settings.backend != "jax":
        return "cpu"

    return runtime_settings.device



from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.core.utilities import pad_z_edges
from veros.core.operators import numpy as npx, update, update_add, update_multiply, at


@veros_kernel
def _calc_cr(rjp, rj, rjm, vel):
    """
    Calculates cr value used in superbee advection scheme
    """
    eps = 1e-20  # prevent division by 0
    return npx.where(vel > 0.0, rjm, rjp) / npx.where(npx.abs(rj) < eps, eps, rj)


@veros_kernel
def limiter(cr):
    return npx.maximum(npx.clip(2 * cr, 0, 1), npx.clip(cr, 0, 2))


@veros_kernel(static_args=("axis"))
def _adv_superbee(state, vel, var, mask, dx, axis):
    vs = state.variables
    settings = state.settings

    if axis == 0:
        sm1, s, sp1, sp2 = ((slice(1 + n, -2 + n or None), slice(2, -2), slice(None)) for n in range(-1, 3))
        dx = vs.cost[npx.newaxis, 2:-2, npx.newaxis] * dx[1:-2, npx.newaxis, npx.newaxis]
    elif axis == 1:
        sm1, s, sp1, sp2 = ((slice(2, -2), slice(1 + n, -2 + n or None), slice(None)) for n in range(-1, 3))
        dx = (vs.cost * dx)[npx.newaxis, 1:-2, npx.newaxis]
    elif axis == 2:
        sm1, s, sp1, sp2 = ((slice(2, -2), slice(2, -2), slice(1 + n, -2 + n or None)) for n in range(-1, 3))
        dx = dx[npx.newaxis, npx.newaxis, :-1]
        vel, var, mask = (pad_z_edges(a) for a in (vel, var, mask))
    else:
        raise ValueError("axis must be 0, 1, or 2")

    rjp = (var[sp2] - var[sp1]) * mask[sp1]
    rj = (var[sp1] - var[s]) * mask[s]
    rjm = (var[s] - var[sm1]) * mask[sm1]
    cr = limiter(_calc_cr(rjp, rj, rjm, vel[s]))

    if axis == 1:
        vel = vel * vs.cosu[npx.newaxis, :, npx.newaxis]

    uCFL = npx.abs(vel[s] * settings.dt_tracer / dx)
    return vel[s] * (var[sp1] + var[s]) * 0.5 - npx.abs(vel[s]) * ((1.0 - cr) + uCFL * cr) * rj * 0.5


@veros_kernel
def adv_flux_2nd(state, var):
    """
    2nd order advective tracer flux
    """
    vs = state.variables

    adv_fe = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_fn = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_ft = allocate(state.dimensions, ("xt", "yt", "zt"))

    adv_fe = update(
        adv_fe,
        at[1:-2, 2:-2, :],
        0.5 * (var[1:-2, 2:-2, :] + var[2:-1, 2:-2, :]) * vs.u[1:-2, 2:-2, :, vs.tau] * vs.maskU[1:-2, 2:-2, :],
    )
    adv_fn = update(
        adv_fn,
        at[2:-2, 1:-2, :],
        vs.cosu[npx.newaxis, 1:-2, npx.newaxis]
        * 0.5
        * (var[2:-2, 1:-2, :] + var[2:-2, 2:-1, :])
        * vs.v[2:-2, 1:-2, :, vs.tau]
        * vs.maskV[2:-2, 1:-2, :],
    )
    adv_ft = update(
        adv_ft,
        at[2:-2, 2:-2, :-1],
        0.5 * (var[2:-2, 2:-2, :-1] + var[2:-2, 2:-2, 1:]) * vs.w[2:-2, 2:-2, :-1, vs.tau] * vs.maskW[2:-2, 2:-2, :-1],
    )
    adv_ft = update(adv_ft, at[:, :, -1], 0.0)

    return adv_fe, adv_fn, adv_ft


@veros_kernel
def adv_flux_superbee(state, var):
    r"""
    from MITgcm
    Calculates advection of a tracer
    using second-order interpolation with a flux limiter:

    \begin{equation*}
    F^x_{adv} = U \overline{ \theta }^i
    - \frac{1}{2} \left([ 1 - \psi(C_r) ] |U|
       + U \frac{u \Delta t}{\Delta x_c} \psi(C_r)
                 \right) \delta_i \theta
    \end{equation*}

    where the $\psi(C_r)$ is the limiter function and $C_r$ is
    the slope ratio.
    """
    vs = state.variables

    adv_fe = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_fn = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_ft = allocate(state.dimensions, ("xt", "yt", "zt"))

    adv_fe = update(adv_fe, at[1:-2, 2:-2, :], _adv_superbee(state, vs.u[..., vs.tau], var, vs.maskU, vs.dxt, 0))
    adv_fn = update(adv_fn, at[2:-2, 1:-2, :], _adv_superbee(state, vs.v[..., vs.tau], var, vs.maskV, vs.dyt, 1))
    adv_ft = update(adv_ft, at[2:-2, 2:-2, :-1], _adv_superbee(state, vs.w[..., vs.tau], var, vs.maskW, vs.dzt, 2))
    adv_ft = update(adv_ft, at[..., -1], 0.0)

    return adv_fe, adv_fn, adv_ft


@veros_routine
def calculate_velocity_on_wgrid(state):
    vs = state.variables
    vs.update(calculate_velocity_on_wgrid_kernel(state))


@veros_kernel
def calculate_velocity_on_wgrid_kernel(state):
    """
    calculates advection velocity for tracer on vs.W grid

    Note: this implementation is not strictly equal to the Fortran version. They only match
    if vs.maskW has exactly one true value across each depth slice.
    """
    vs = state.variables

    # lateral advection velocities on W grid
    vs.u_wgrid = update(
        vs.u_wgrid,
        at[:, :, :-1],
        vs.u[:, :, 1:, vs.tau]
        * vs.maskU[:, :, 1:]
        * 0.5
        * vs.dzt[npx.newaxis, npx.newaxis, 1:]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        + vs.u[:, :, :-1, vs.tau]
        * vs.maskU[:, :, :-1]
        * 0.5
        * vs.dzt[npx.newaxis, npx.newaxis, :-1]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )
    vs.v_wgrid = update(
        vs.v_wgrid,
        at[:, :, :-1],
        vs.v[:, :, 1:, vs.tau]
        * vs.maskV[:, :, 1:]
        * 0.5
        * vs.dzt[npx.newaxis, npx.newaxis, 1:]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        + vs.v[:, :, :-1, vs.tau]
        * vs.maskV[:, :, :-1]
        * 0.5
        * vs.dzt[npx.newaxis, npx.newaxis, :-1]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )
    vs.u_wgrid = update(
        vs.u_wgrid, at[:, :, -1], vs.u[:, :, -1, vs.tau] * vs.maskU[:, :, -1] * 0.5 * vs.dzt[-1:] / vs.dzw[-1:]
    )
    vs.v_wgrid = update(
        vs.v_wgrid, at[:, :, -1], vs.v[:, :, -1, vs.tau] * vs.maskV[:, :, -1] * 0.5 * vs.dzt[-1:] / vs.dzw[-1:]
    )

    # redirect velocity at bottom and at topography
    vs.u_wgrid = update(
        vs.u_wgrid,
        at[:, :, 0],
        vs.u_wgrid[:, :, 0] + vs.u[:, :, 0, vs.tau] * vs.maskU[:, :, 0] * 0.5 * vs.dzt[0] / vs.dzw[0],
    )
    vs.v_wgrid = update(
        vs.v_wgrid,
        at[:, :, 0],
        vs.v_wgrid[:, :, 0] + vs.v[:, :, 0, vs.tau] * vs.maskV[:, :, 0] * 0.5 * vs.dzt[0] / vs.dzw[0],
    )
    mask = vs.maskW[:-1, :, :-1] * vs.maskW[1:, :, :-1]
    vs.u_wgrid = update_add(
        vs.u_wgrid,
        at[:-1, :, 1:],
        (vs.u_wgrid[:-1, :, :-1] * vs.dzw[npx.newaxis, npx.newaxis, :-1] / vs.dzw[npx.newaxis, npx.newaxis, 1:])
        * (1.0 - mask),
    )
    vs.u_wgrid = update_multiply(vs.u_wgrid, at[:-1, :, :-1], mask)
    mask = vs.maskW[:, :-1, :-1] * vs.maskW[:, 1:, :-1]
    vs.v_wgrid = update_add(
        vs.v_wgrid,
        at[:, :-1, 1:],
        (vs.v_wgrid[:, :-1, :-1] * vs.dzw[npx.newaxis, npx.newaxis, :-1] / vs.dzw[npx.newaxis, npx.newaxis, 1:])
        * (1.0 - mask),
    )
    vs.v_wgrid = update_multiply(vs.v_wgrid, at[:, :-1, :-1], mask)

    # vertical advection velocity on W grid from continuity
    vs.w_wgrid = update(vs.w_wgrid, at[:, :, 0], 0.0)
    vs.w_wgrid = update(
        vs.w_wgrid,
        at[1:, 1:, :],
        npx.cumsum(
            -vs.dzw[npx.newaxis, npx.newaxis, :]
            * (
                (vs.u_wgrid[1:, 1:, :] - vs.u_wgrid[:-1, 1:, :])
                / (vs.cost[npx.newaxis, 1:, npx.newaxis] * vs.dxt[1:, npx.newaxis, npx.newaxis])
                + (
                    vs.cosu[npx.newaxis, 1:, npx.newaxis] * vs.v_wgrid[1:, 1:, :]
                    - vs.cosu[npx.newaxis, :-1, npx.newaxis] * vs.v_wgrid[1:, :-1, :]
                )
                / (vs.cost[npx.newaxis, 1:, npx.newaxis] * vs.dyt[npx.newaxis, 1:, npx.newaxis])
            ),
            axis=2,
        ),
    )

    return KernelOutput(u_wgrid=vs.u_wgrid, v_wgrid=vs.v_wgrid, w_wgrid=vs.w_wgrid)


@veros_kernel
def adv_flux_superbee_wgrid(state, var):
    """
    Calculates advection of a tracer defined on Wgrid
    """
    vs = state.variables

    adv_fe = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_fn = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_ft = allocate(state.dimensions, ("xt", "yt", "zt"))

    maskUtr = allocate(state.dimensions, ("xt", "yt", "zw"))
    maskUtr = update(maskUtr, at[:-1, :, :], vs.maskW[1:, :, :] * vs.maskW[:-1, :, :])
    adv_fe = update(adv_fe, at[1:-2, 2:-2, :], _adv_superbee(state, vs.u_wgrid, var, maskUtr, vs.dxt, axis=0))

    maskVtr = allocate(state.dimensions, ("xt", "yt", "zw"))
    maskVtr = update(maskVtr, at[:, :-1, :], vs.maskW[:, 1:, :] * vs.maskW[:, :-1, :])
    adv_fn = update(adv_fn, at[2:-2, 1:-2, :], _adv_superbee(state, vs.v_wgrid, var, maskVtr, vs.dyt, axis=1))

    maskWtr = allocate(state.dimensions, ("xt", "yt", "zw"))
    maskWtr = update(maskWtr, at[:, :, :-1], vs.maskW[:, :, 1:] * vs.maskW[:, :, :-1])
    adv_ft = update(adv_ft, at[2:-2, 2:-2, :-1], _adv_superbee(state, vs.w_wgrid, var, maskWtr, vs.dzw, axis=2))
    adv_ft = update(adv_ft, at[..., -1], 0.0)

    return adv_fe, adv_fn, adv_ft


@veros_kernel
def adv_flux_upwind_wgrid(state, var):
    """
    Calculates advection of a tracer defined on Wgrid
    """
    vs = state.variables

    adv_fe = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_fn = allocate(state.dimensions, ("xt", "yt", "zt"))
    adv_ft = allocate(state.dimensions, ("xt", "yt", "zt"))

    maskUtr = vs.maskW[2:-1, 2:-2, :] * vs.maskW[1:-2, 2:-2, :]
    rj = (var[2:-1, 2:-2, :] - var[1:-2, 2:-2, :]) * maskUtr
    adv_fe = update(
        adv_fe,
        at[1:-2, 2:-2, :],
        vs.u_wgrid[1:-2, 2:-2, :] * (var[2:-1, 2:-2, :] + var[1:-2, 2:-2, :]) * 0.5
        - npx.abs(vs.u_wgrid[1:-2, 2:-2, :]) * rj * 0.5,
    )

    maskVtr = vs.maskW[2:-2, 2:-1, :] * vs.maskW[2:-2, 1:-2, :]
    rj = (var[2:-2, 2:-1, :] - var[2:-2, 1:-2, :]) * maskVtr
    adv_fn = update(
        adv_fn,
        at[2:-2, 1:-2, :],
        vs.cosu[npx.newaxis, 1:-2, npx.newaxis]
        * vs.v_wgrid[2:-2, 1:-2, :]
        * (var[2:-2, 2:-1, :] + var[2:-2, 1:-2, :])
        * 0.5
        - npx.abs(vs.cosu[npx.newaxis, 1:-2, npx.newaxis] * vs.v_wgrid[2:-2, 1:-2, :]) * rj * 0.5,
    )

    maskWtr = vs.maskW[2:-2, 2:-2, 1:] * vs.maskW[2:-2, 2:-2, :-1]
    rj = (var[2:-2, 2:-2, 1:] - var[2:-2, 2:-2, :-1]) * maskWtr
    adv_ft = update(
        adv_ft,
        at[2:-2, 2:-2, :-1],
        vs.w_wgrid[2:-2, 2:-2, :-1] * (var[2:-2, 2:-2, 1:] + var[2:-2, 2:-2, :-1]) * 0.5
        - npx.abs(vs.w_wgrid[2:-2, 2:-2, :-1]) * rj * 0.5,
    )
    adv_ft = update(adv_ft, at[:, :, -1], 0.0)

    return adv_fe, adv_fn, adv_ft



from veros.core.operators import numpy as npx

from veros import veros_kernel, KernelOutput
from veros.variables import allocate
from veros.core import numerics, utilities
from veros.core.operators import update, update_add, at


@veros_kernel
def isoneutral_friction(state):
    """
    vertical friction using TEM formalism for eddy driven velocity
    """
    vs = state.variables
    settings = state.settings

    flux_top = allocate(state.dimensions, ("xt", "yt", "zt"))
    delta, a_tri, b_tri, c_tri = (allocate(state.dimensions, ("xt", "yt", "zt"))[1:-2, 1:-2, :] for _ in range(4))

    if settings.enable_implicit_vert_friction:
        aloc = vs.u[:, :, :, vs.taup1]
    else:
        aloc = vs.u[:, :, :, vs.tau]

    # implicit vertical friction of zonal momentum by GM
    ks = npx.maximum(vs.kbot[1:-2, 1:-2], vs.kbot[2:-1, 1:-2])
    _, water_mask, edge_mask = utilities.create_water_masks(ks, settings.nz)

    fxa = 0.5 * (vs.kappa_gm[1:-2, 1:-2, :] + vs.kappa_gm[2:-1, 1:-2, :])
    delta = update(
        delta,
        at[:, :, :-1],
        settings.dt_mom
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        * fxa[:, :, :-1]
        * vs.maskU[1:-2, 1:-2, 1:]
        * vs.maskU[1:-2, 1:-2, :-1],
    )
    delta = update(delta, at[..., -1], 0.0)
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri_edge = 1 + delta / vs.dzt[npx.newaxis, npx.newaxis, :]
    b_tri = update(
        b_tri,
        at[:, :, 1:-1],
        1
        + delta[:, :, 1:-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1]
        + delta[:, :, :-2] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1],
    )
    b_tri = update(b_tri, at[:, :, -1], 1 + delta[:, :, -2] / vs.dzt[-1])
    c_tri = update(c_tri, at[...], -delta / vs.dzt[npx.newaxis, npx.newaxis, :])

    sol = utilities.solve_implicit(
        a_tri, b_tri, c_tri, aloc[1:-2, 1:-2, :], water_mask, b_edge=b_tri_edge, edge_mask=edge_mask
    )
    vs.u = update(vs.u, at[1:-2, 1:-2, :, vs.taup1], npx.where(water_mask, sol, vs.u[1:-2, 1:-2, :, vs.taup1]))
    vs.du_mix = update_add(
        vs.du_mix,
        at[1:-2, 1:-2, :],
        (vs.u[1:-2, 1:-2, :, vs.taup1] - aloc[1:-2, 1:-2, :]) / settings.dt_mom * water_mask,
    )

    if settings.enable_conserve_energy:
        # diagnose dissipation
        diss = allocate(state.dimensions, ("xt", "yt", "zt"))
        fxa = 0.5 * (vs.kappa_gm[1:-2, 1:-2, :-1] + vs.kappa_gm[2:-1, 1:-2, :-1])
        flux_top = update(
            flux_top,
            at[1:-2, 1:-2, :-1],
            fxa
            * (vs.u[1:-2, 1:-2, 1:, vs.taup1] - vs.u[1:-2, 1:-2, :-1, vs.taup1])
            / vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskU[1:-2, 1:-2, 1:]
            * vs.maskU[1:-2, 1:-2, :-1],
        )
        diss = update(
            diss,
            at[1:-2, 1:-2, :-1],
            (vs.u[1:-2, 1:-2, 1:, vs.tau] - vs.u[1:-2, 1:-2, :-1, vs.tau])
            * flux_top[1:-2, 1:-2, :-1]
            / vs.dzw[npx.newaxis, npx.newaxis, :-1],
        )
        diss = update(diss, at[:, :, -1], 0.0)
        diss = numerics.ugrid_to_tgrid(state, diss)
        vs.K_diss_gm = diss

    if settings.enable_implicit_vert_friction:
        aloc = vs.v[:, :, :, vs.taup1]
    else:
        aloc = vs.v[:, :, :, vs.tau]

    # implicit vertical friction of zonal momentum by GM
    ks = npx.maximum(vs.kbot[1:-2, 1:-2], vs.kbot[1:-2, 2:-1])
    _, water_mask, edge_mask = utilities.create_water_masks(ks, settings.nz)

    fxa = 0.5 * (vs.kappa_gm[1:-2, 1:-2, :] + vs.kappa_gm[1:-2, 2:-1, :])
    delta = update(
        delta,
        at[:, :, :-1],
        settings.dt_mom
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        * fxa[:, :, :-1]
        * vs.maskV[1:-2, 1:-2, 1:]
        * vs.maskV[1:-2, 1:-2, :-1],
    )
    delta = update(delta, at[..., -1], 0.0)
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri_edge = 1 + delta / vs.dzt[npx.newaxis, npx.newaxis, :]
    b_tri = update(
        b_tri,
        at[:, :, 1:-1],
        1
        + delta[:, :, 1:-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1]
        + delta[:, :, :-2] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1],
    )
    b_tri = update(b_tri, at[:, :, -1], 1 + delta[:, :, -2] / vs.dzt[-1])
    c_tri = update(c_tri, at[...], -delta / vs.dzt[npx.newaxis, npx.newaxis, :])

    sol = utilities.solve_implicit(
        a_tri, b_tri, c_tri, aloc[1:-2, 1:-2, :], water_mask, b_edge=b_tri_edge, edge_mask=edge_mask
    )
    vs.v = update(vs.v, at[1:-2, 1:-2, :, vs.taup1], npx.where(water_mask, sol, vs.v[1:-2, 1:-2, :, vs.taup1]))
    vs.dv_mix = update_add(
        vs.dv_mix,
        at[1:-2, 1:-2, :],
        (vs.v[1:-2, 1:-2, :, vs.taup1] - aloc[1:-2, 1:-2, :]) / settings.dt_mom * water_mask,
    )

    if settings.enable_conserve_energy:
        # diagnose dissipation
        diss = allocate(state.dimensions, ("xt", "yt", "zt"))
        fxa = 0.5 * (vs.kappa_gm[1:-2, 1:-2, :-1] + vs.kappa_gm[1:-2, 2:-1, :-1])
        flux_top = update(
            flux_top,
            at[1:-2, 1:-2, :-1],
            fxa
            * (vs.v[1:-2, 1:-2, 1:, vs.taup1] - vs.v[1:-2, 1:-2, :-1, vs.taup1])
            / vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskV[1:-2, 1:-2, 1:]
            * vs.maskV[1:-2, 1:-2, :-1],
        )
        diss = update(
            diss,
            at[1:-2, 1:-2, :-1],
            (vs.v[1:-2, 1:-2, 1:, vs.tau] - vs.v[1:-2, 1:-2, :-1, vs.tau])
            * flux_top[1:-2, 1:-2, :-1]
            / vs.dzw[npx.newaxis, npx.newaxis, :-1],
        )
        diss = update(diss, at[:, :, -1], 0.0)
        diss = numerics.vgrid_to_tgrid(state, diss)
        vs.K_diss_gm = vs.K_diss_gm + diss

    return KernelOutput(u=vs.u, v=vs.v, du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_gm=vs.K_diss_gm)



try:
    from veros.core.special import tdma_cython_
except ImportError:
    HAS_CPU_EXT = False
else:
    HAS_CPU_EXT = True

try:
    from veros.core.special import tdma_cuda_
except ImportError:
    HAS_GPU_EXT = False
else:
    HAS_GPU_EXT = True

import numpy as np
import jax.numpy as jnp

import jax
from jax import abstract_arrays
from jax.core import Primitive
from jax.lib import xla_client
from jax.interpreters import xla

_ops = xla_client.ops

if HAS_CPU_EXT:
    for kernel_name in (b"tdma_cython_double", b"tdma_cython_float"):
        fn = tdma_cython_.cpu_custom_call_targets[kernel_name]
        xla_client.register_custom_call_target(kernel_name, fn, platform="cpu")

if HAS_GPU_EXT:
    for kernel_name in (b"tdma_cuda_double", b"tdma_cuda_float"):
        fn = tdma_cuda_.gpu_custom_call_targets[kernel_name]
        xla_client.register_custom_call_target(kernel_name, fn, platform="gpu")


def _constant_s64_scalar(c, x):
    return _ops.Constant(c, np.int64(x))


def tdma(a, b, c, d, interior_mask, edge_mask, device=None):
    if device is None:
        device = jax.default_backend()

    if not a.shape == b.shape == c.shape == d.shape:
        raise ValueError("all inputs must have identical shape")

    if not a.dtype == b.dtype == c.dtype == d.dtype:
        raise ValueError("all inputs must have the same dtype")

    if device == "cpu":
        system_depths = jnp.sum(interior_mask, axis=-1, dtype="int32")
        return tdma_p.bind(a, b, c, d, system_depths)

    a = interior_mask * a * jnp.logical_not(edge_mask)
    b = jnp.where(interior_mask, b, 1.0)
    c = interior_mask * c
    d = interior_mask * d

    return tdma_p.bind(a, b, c, d, system_depths=None)


def tdma_impl(*args, **kwargs):
    return xla.apply_primitive(tdma_p, *args, **kwargs)


def tdma_xla_encode_cpu(builder, a, b, c, d, system_depths):
    # try import again to trigger exception on ImportError
    from veros.core.special import tdma_cython_  # noqa: F401

    x_shape = builder.GetShape(a)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    supported_dtypes = (
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    if dtype not in supported_dtypes:
        raise TypeError(f"TDMA only supports {supported_dtypes} arrays, got: {dtype}")

    # compute number of elements to vectorize over
    num_systems = 1
    for el in dims[:-1]:
        num_systems *= el

    stride = dims[-1]

    sys_depth_shape = builder.get_shape(system_depths)
    sys_depth_dtype = sys_depth_shape.element_type()
    sys_depth_dims = sys_depth_shape.dimensions()
    assert sys_depth_dtype is np.dtype(np.int32)
    assert tuple(sys_depth_dims) == tuple(dims[:-1])

    arr_shape = xla_client.Shape.array_shape(dtype, dims)
    out_shape = xla_client.Shape.tuple_shape([arr_shape, xla_client.Shape.array_shape(dtype, (stride,))])

    if dtype is np.dtype(np.float32):
        kernel = b"tdma_cython_float"
    elif dtype is np.dtype(np.float64):
        kernel = b"tdma_cython_double"
    else:
        raise RuntimeError("got unrecognized dtype")

    out = _ops.CustomCall(
        builder,
        kernel,
        operands=(
            a,
            b,
            c,
            d,
            system_depths,
            _constant_s64_scalar(builder, num_systems),
            _constant_s64_scalar(builder, stride),
        ),
        shape=out_shape,
    )
    return _ops.GetTupleElement(out, 0)


def tdma_xla_encode_gpu(builder, a, b, c, d, system_depths):
    # try import again to trigger exception on ImportError
    from veros.core.special import tdma_cuda_  # noqa: F401

    if system_depths is not None:
        raise ValueError("TDMA does not support system_depths argument on GPU")

    a_shape = builder.get_shape(a)
    dtype = a_shape.element_type()
    dims = a_shape.dimensions()

    supported_dtypes = (
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    if dtype not in supported_dtypes:
        raise TypeError(f"TDMA only supports {supported_dtypes} arrays, got: {dtype}")

    # compute number of elements to vectorize over
    num_systems = 1
    for el in dims[:-1]:
        num_systems *= el

    system_depth = dims[-1]

    if dtype is np.dtype(np.float32):
        kernel = b"tdma_cuda_float"
    elif dtype is np.dtype(np.float64):
        kernel = b"tdma_cuda_double"
    else:
        raise RuntimeError("got unrecognized dtype")

    opaque = tdma_cuda_.build_tridiag_descriptor(num_systems, system_depth)

    ndims = len(dims)
    arr_layout = tuple(range(ndims - 2, -1, -1)) + (ndims - 1,)
    arr_shape = xla_client.Shape.array_shape(dtype, dims, arr_layout)
    out_shape = xla_client.Shape.tuple_shape([arr_shape, arr_shape])

    out = _ops.CustomCallWithLayout(
        builder,
        kernel,
        operands=(a, b, c, d),
        shape_with_layout=out_shape,
        operand_shapes_with_layout=(arr_shape,) * 4,
        opaque=opaque,
    )
    return _ops.GetTupleElement(out, 0)


def tdma_abstract_eval(a, b, c, d, system_depths):
    return abstract_arrays.ShapedArray(a.shape, a.dtype)


tdma_p = Primitive("tdma")
tdma_p.def_impl(tdma_impl)
tdma_p.def_abstract_eval(tdma_abstract_eval)

xla.backend_specific_translations["cpu"][tdma_p] = tdma_xla_encode_cpu
xla.backend_specific_translations["gpu"][tdma_p] = tdma_xla_encode_gpu



import os
from threading import local
from collections import namedtuple

from veros.backend import BACKENDS
from veros.logs import LOGLEVELS


# globals
log_args = local()
log_args.log_all_processes = False
log_args.loglevel = "info"


# MPI helpers


def _default_mpi_comm():
    try:
        from mpi4py import MPI
    except ImportError:
        return None
    else:
        return MPI.COMM_WORLD


# validators


def parse_two_ints(v):
    return (int(v[0]), int(v[1]))


def parse_choice(choices, preserve_case=False):
    def validate(choice):
        if isinstance(choice, str) and not preserve_case:
            choice = choice.lower()

        if choice not in choices:
            raise ValueError(f"must be one of {choices}")

        return choice

    return validate


def parse_bool(obj):
    if not isinstance(obj, str):
        return bool(obj)

    return obj.lower() in {"1", "true", "on"}


def check_mpi_comm(comm):
    if comm is not None:
        from mpi4py import MPI

        if not isinstance(comm, MPI.Comm):
            raise TypeError("mpi_comm must be Comm instance or None")

    return comm


def set_loglevel(val):
    from veros import logs

    log_args.loglevel = parse_choice(LOGLEVELS)(val)
    logs.setup_logging(loglevel=log_args.loglevel, log_all_processes=log_args.log_all_processes)
    return log_args.loglevel


def set_log_all_processes(val):
    from veros import logs

    log_args.log_all_processes = parse_bool(val)
    logs.setup_logging(loglevel=log_args.loglevel, log_all_processes=log_args.log_all_processes)
    return log_args.log_all_processes


DEVICES = ("cpu", "gpu", "tpu")
FLOAT_TYPES = ("float64", "float32")
LINEAR_SOLVERS = ("scipy", "scipy_jax", "petsc", "best")


# settings

RuntimeSetting = namedtuple("RuntimeSetting", ("type", "default", "read_from_env"))
RuntimeSetting.__new__.__defaults__ = (None, None, True)

AVAILABLE_SETTINGS = {
    "backend": RuntimeSetting(parse_choice(BACKENDS), "numpy"),
    "device": RuntimeSetting(parse_choice(DEVICES), "cpu"),
    "float_type": RuntimeSetting(parse_choice(FLOAT_TYPES), "float64"),
    "linear_solver": RuntimeSetting(parse_choice(LINEAR_SOLVERS), "best"),
    "petsc_options": RuntimeSetting(str, ""),
    "monitor_streamfunction_residual": RuntimeSetting(parse_bool, True),
    "num_proc": RuntimeSetting(parse_two_ints, (1, 1), read_from_env=False),
    "profile_mode": RuntimeSetting(parse_bool, False),
    "loglevel": RuntimeSetting(set_loglevel, "info"),
    "mpi_comm": RuntimeSetting(check_mpi_comm, _default_mpi_comm(), read_from_env=False),
    "log_all_processes": RuntimeSetting(set_log_all_processes, False),
    "use_io_threads": RuntimeSetting(parse_bool, False),
    "io_timeout": RuntimeSetting(float, 20),
    "hdf5_gzip_compression": RuntimeSetting(bool, True),
    "force_overwrite": RuntimeSetting(bool, False),
    "diskless_mode": RuntimeSetting(bool, False),
    "pyom_compatibility_mode": RuntimeSetting(bool, False),
}


class RuntimeSettings:
    __slots__ = ["__locked__", "__setting_types__", "__settings__", *AVAILABLE_SETTINGS.keys()]

    def __init__(self, **kwargs):
        self.__locked__ = False
        self.__setting_types__ = {}

        for name, setting in AVAILABLE_SETTINGS.items():
            setting_envvar = f"VEROS_{name.upper()}"

            if name in kwargs:
                val = kwargs[name]
            elif setting.read_from_env:
                val = os.environ.get(setting_envvar, setting.default)
            else:
                val = setting.default

            self.__setting_types__[name] = setting.type
            self.__setattr__(name, val)

        self.__settings__ = set(self.__setting_types__.keys())

    def update(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

        return self

    def __setattr__(self, attr, val):
        if getattr(self, "__locked__", False):
            raise RuntimeError("Runtime settings cannot be modified after import of core modules")

        if attr.startswith("_"):
            return super().__setattr__(attr, val)

        # coerce type
        stype = self.__setting_types__.get(attr)
        if stype is not None:
            try:
                val = stype(val)
            except (TypeError, ValueError) as e:
                raise ValueError(f'Got invalid value for runtime setting "{attr}": {e!s}') from None

        return super().__setattr__(attr, val)

    def __repr__(self):
        setval = ", ".join(f"{key}={repr(getattr(self, key))}" for key in self.__settings__)
        return f"{self.__class__.__name__}({setval})"


# state


class RuntimeState:
    """Unifies attributes from various modules in a simple read-only object"""

    __slots__ = ()

    @property
    def proc_rank(self):
        from veros import runtime_settings

        comm = runtime_settings.mpi_comm

        if comm is None:
            return 0

        return comm.Get_rank()

    @property
    def proc_num(self):
        from veros import runtime_settings

        comm = runtime_settings.mpi_comm

        if comm is None:
            return 1

        return comm.Get_size()

    @property
    def proc_idx(self):
        from veros import distributed

        return distributed.proc_rank_to_index(self.proc_rank)

    @property
    def backend_module(self):
        from veros import backend, runtime_settings

        return backend.get_backend_module(runtime_settings.backend)

    def __setattr__(self, attr, val):
        raise TypeError(f"Cannot modify {self.__class__.__name__} objects")



from veros.core.operators import numpy as npx

from veros import veros_kernel
from veros.core.operators import update, at, solve_tridiagonal


@veros_kernel(static_args=("enable_cyclic_x", "local"))
def enforce_boundaries(arr, enable_cyclic_x, local=False):
    from veros import runtime_state as rst
    from veros.routines import CURRENT_CONTEXT

    if rst.proc_num == 1 or not CURRENT_CONTEXT.is_dist_safe or local:
        if enable_cyclic_x:
            arr = update(arr, at[-2:, ...], arr[2:4, ...])
            arr = update(arr, at[:2, ...], arr[-4:-2, ...])
        return arr

    from veros.distributed import exchange_overlap

    arr = exchange_overlap(arr, ["xt", "yt"], cyclic=enable_cyclic_x)
    return arr


@veros_kernel
def pad_z_edges(array):
    """
    Pads the z-axis of an array by repeating its edge values
    """
    if array.ndim == 1:
        newarray = npx.pad(array, 1, mode="edge")
    elif array.ndim >= 3:
        newarray = npx.pad(array, ((0, 0), (0, 0), (1, 1)), mode="edge")
    else:
        raise ValueError("Array to pad needs to have 1 or at least 3 dimensions")
    return newarray


@veros_kernel(static_args=("nz"))
def create_water_masks(ks, nz):
    ks = ks - 1
    land_mask = ks >= 0
    water_mask = npx.logical_and(
        land_mask[:, :, npx.newaxis], npx.arange(nz)[npx.newaxis, npx.newaxis, :] >= ks[:, :, npx.newaxis]
    )
    edge_mask = npx.logical_and(
        land_mask[:, :, npx.newaxis], npx.arange(nz)[npx.newaxis, npx.newaxis, :] == ks[:, :, npx.newaxis]
    )
    return land_mask, water_mask, edge_mask


@veros_kernel
def solve_implicit(a, b, c, d, water_mask, edge_mask, b_edge=None, d_edge=None):
    if b_edge is not None:
        b = npx.where(edge_mask, b_edge, b)

    if d_edge is not None:
        d = npx.where(edge_mask, d_edge, d)

    return solve_tridiagonal(a, b, c, d, water_mask, edge_mask)



#!/usr/bin/env python

import os

import h5netcdf
from PIL import Image
import scipy.spatial
import scipy.ndimage

from veros import VerosSetup, veros_routine, veros_kernel, KernelOutput
from veros.variables import Variable
from veros.core.operators import numpy as npx, update, at
import veros.tools

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_FILES = veros.tools.get_assets("north_atlantic", os.path.join(BASE_PATH, "assets.json"))
TOPO_MASK_FILE = os.path.join(BASE_PATH, "topo_mask.png")


class NorthAtlanticSetup(VerosSetup):
    """A regional model of the North Atlantic, inspired by `Smith et al., 2000`_.

    Forcing and initial conditions are taken from the FLAME PyOM2 setup. Bathymetry
    data from ETOPO1 (resolution of 1 arcmin).

    Boundary forcings are implemented via sponge layers in the Greenland Sea, by the
    Strait of Gibraltar, and in the South Atlantic. This setup runs with arbitrary resolution;
    upon changing the number of grid cells, all forcing files will be interpolated to
    the new grid. Default resolution corresponds roughly to :math:`0.5 \\times 0.25` degrees.

    .. _Smith et al., 2000:
       http://journals.ametsoc.org/doi/10.1175/1520-0485%282000%29030%3C1532%3ANSOTNA%3E2.0.CO%3B2
    """

    x_boundary = 17.2
    y_boundary = 70.0
    max_depth = 5800.0

    @veros_routine
    def set_parameter(self, state):
        settings = state.settings

        settings.identifier = "na"

        settings.nx, settings.ny, settings.nz = 250, 350, 50
        settings.x_origin = -98.0
        settings.y_origin = -18.0

        settings.dt_mom = 3600.0 / 2.0
        settings.dt_tracer = 3600.0 / 2.0
        settings.runlen = 86400 * 365.0 * 10.0

        settings.coord_degree = True

        settings.enable_neutral_diffusion = True
        settings.enable_skew_diffusion = True
        settings.K_iso_0 = 1000.0
        settings.K_iso_steep = 200.0
        settings.iso_dslope = 1.0 / 1000.0
        settings.iso_slopec = 4.0 / 1000.0

        settings.enable_hor_friction = True
        settings.A_h = 1e3
        settings.enable_hor_friction_cos_scaling = True
        settings.hor_friction_cosPower = 1
        settings.enable_tempsalt_sources = True

        settings.enable_implicit_vert_friction = True
        settings.enable_tke = True
        settings.c_k = 0.1
        settings.c_eps = 0.7
        settings.alpha_tke = 30.0
        settings.mxl_min = 1e-8
        settings.tke_mxl_choice = 2
        settings.kappaM_min = 2e-4
        settings.kappaH_min = 2e-5
        settings.enable_kappaH_profile = True

        settings.K_gm_0 = 1000.0

        settings.enable_eke = False
        settings.enable_idemix = False

        settings.eq_of_state_type = 5

        state.dimensions["nmonths"] = 12
        state.var_meta.update(
            {
                "sss_clim": Variable("sss_clim", ("xt", "yt", "nmonths"), "g/kg", "Monthly sea surface salinity"),
                "sst_clim": Variable("sst_clim", ("xt", "yt", "nmonths"), "deg C", "Monthly sea surface temperature"),
                "sss_rest": Variable(
                    "sss_rest", ("xt", "yt", "nmonths"), "g/kg", "Monthly sea surface salinity restoring"
                ),
                "sst_rest": Variable(
                    "sst_rest", ("xt", "yt", "nmonths"), "deg C", "Monthly sea surface temperature restoring"
                ),
                "t_star": Variable(
                    "t_star", ("xt", "yt", "zt", "nmonths"), "deg C", "Temperature sponge layer forcing"
                ),
                "s_star": Variable("s_star", ("xt", "yt", "zt", "nmonths"), "g/kg", "Salinity sponge layer forcing"),
                "rest_tscl": Variable("rest_tscl", ("xt", "yt", "zt"), "1/s", "Forcing restoration time scale"),
                "taux": Variable("taux", ("xt", "yt", "nmonths"), "N/s^2", "Monthly zonal wind stress"),
                "tauy": Variable("tauy", ("xt", "yt", "nmonths"), "N/s^2", "Monthly meridional wind stress"),
            }
        )

    @veros_routine
    def set_grid(self, state):
        vs = state.variables
        settings = state.settings

        vs.dxt = update(vs.dxt, at[2:-2], (self.x_boundary - settings.x_origin) / settings.nx)
        vs.dyt = update(vs.dyt, at[2:-2], (self.y_boundary - settings.y_origin) / settings.ny)
        vs.dzt = veros.tools.get_vinokur_grid_steps(settings.nz, self.max_depth, 10.0, refine_towards="lower")

    @veros_routine
    def set_coriolis(self, state):
        vs = state.variables
        settings = state.settings
        vs.coriolis_t = update(
            vs.coriolis_t, at[...], 2 * settings.omega * npx.sin(vs.yt[npx.newaxis, :] / 180.0 * settings.pi)
        )

    @veros_routine(dist_safe=False, local_variables=["kbot", "xt", "yt", "zt"])
    def set_topography(self, state):
        vs = state.variables
        settings = state.settings

        with h5netcdf.File(DATA_FILES["topography"], "r") as topo_file:
            topo_x, topo_y, topo_bottom_depth = (self._get_data(topo_file, k) for k in ("x", "y", "z"))

        topo_mask = npx.flipud(npx.asarray(Image.open(TOPO_MASK_FILE))).T
        topo_bottom_depth = npx.where(topo_mask, 0, topo_bottom_depth)
        topo_bottom_depth = scipy.ndimage.gaussian_filter(
            topo_bottom_depth, sigma=(len(topo_x) / settings.nx, len(topo_y) / settings.ny)
        )
        interp_coords = npx.meshgrid(vs.xt[2:-2], vs.yt[2:-2], indexing="ij")
        interp_coords = npx.rollaxis(npx.asarray(interp_coords), 0, 3)
        z_interp = scipy.interpolate.interpn(
            (topo_x, topo_y), topo_bottom_depth, interp_coords, method="nearest", bounds_error=False, fill_value=0
        )
        vs.kbot = update(
            vs.kbot,
            at[2:-2, 2:-2],
            npx.where(
                z_interp < 0.0,
                1 + npx.argmin(npx.abs(z_interp[:, :, npx.newaxis] - vs.zt[npx.newaxis, npx.newaxis, :]), axis=2),
                0,
            ),
        )
        vs.kbot = npx.where(vs.kbot < settings.nz, vs.kbot, 0)

    def _get_data(self, f, var):
        """Retrieve variable from h5netcdf file"""
        var_obj = f.variables[var]
        return npx.array(var_obj).T

    @veros_routine(
        dist_safe=False,
        local_variables=[
            "tau",
            "xt",
            "yt",
            "zt",
            "temp",
            "maskT",
            "salt",
            "taux",
            "tauy",
            "sst_clim",
            "sss_clim",
            "sst_rest",
            "sss_rest",
            "t_star",
            "s_star",
            "rest_tscl",
        ],
    )
    def set_initial_conditions(self, state):
        vs = state.variables

        with h5netcdf.File(DATA_FILES["forcing"], "r") as forcing_file:
            t_hor = (vs.xt[2:-2], vs.yt[2:-2])
            t_grid = (vs.xt[2:-2], vs.yt[2:-2], vs.zt)

            forc_coords = [self._get_data(forcing_file, k) for k in ("xt", "yt", "zt")]
            forc_coords[0] = forc_coords[0] - 360
            forc_coords[2] = -0.01 * forc_coords[2][::-1]

            temp_raw = self._get_data(forcing_file, "temp_ic")[..., ::-1]
            temp = veros.tools.interpolate(forc_coords, temp_raw, t_grid, missing_value=-1e20)
            vs.temp = update(vs.temp, at[2:-2, 2:-2, :, vs.tau], vs.maskT[2:-2, 2:-2, :] * temp)

            salt_raw = self._get_data(forcing_file, "salt_ic")[..., ::-1]
            salt = 35.0 + 1000 * veros.tools.interpolate(forc_coords, salt_raw, t_grid, missing_value=-1e20)
            vs.salt = update(vs.salt, at[2:-2, 2:-2, :, vs.tau], vs.maskT[2:-2, 2:-2, :] * salt)

            forc_u_coords_hor = [self._get_data(forcing_file, k) for k in ("xu", "yu")]
            forc_u_coords_hor[0] = forc_u_coords_hor[0] - 360

            taux = self._get_data(forcing_file, "taux")
            tauy = self._get_data(forcing_file, "tauy")
            for k in range(12):
                vs.taux = update(
                    vs.taux,
                    at[2:-2, 2:-2, k],
                    (veros.tools.interpolate(forc_u_coords_hor, taux[..., k], t_hor, missing_value=-1e20) / 10.0),
                )
                vs.tauy = update(
                    vs.tauy,
                    at[2:-2, 2:-2, k],
                    (veros.tools.interpolate(forc_u_coords_hor, tauy[..., k], t_hor, missing_value=-1e20) / 10.0),
                )

            # heat flux and salinity restoring

            sst_clim, sss_clim, sst_rest, sss_rest = [
                forcing_file.variables[k][...].T for k in ("sst_clim", "sss_clim", "sst_rest", "sss_rest")
            ]

        for k in range(12):
            vs.sst_clim = update(
                vs.sst_clim,
                at[2:-2, 2:-2, k],
                veros.tools.interpolate(forc_coords[:-1], sst_clim[..., k], t_hor, missing_value=-1e20),
            )
            vs.sss_clim = update(
                vs.sss_clim,
                at[2:-2, 2:-2, k],
                (veros.tools.interpolate(forc_coords[:-1], sss_clim[..., k], t_hor, missing_value=-1e20) * 1000 + 35),
            )
            vs.sst_rest = update(
                vs.sst_rest,
                at[2:-2, 2:-2, k],
                (veros.tools.interpolate(forc_coords[:-1], sst_rest[..., k], t_hor, missing_value=-1e20) * 41868.0),
            )
            vs.sss_rest = update(
                vs.sss_rest,
                at[2:-2, 2:-2, k],
                (veros.tools.interpolate(forc_coords[:-1], sss_rest[..., k], t_hor, missing_value=-1e20) / 100.0),
            )

        with h5netcdf.File(DATA_FILES["restoring"], "r") as restoring_file:
            rest_coords = [self._get_data(restoring_file, k) for k in ("xt", "yt", "zt")]
            rest_coords[0] = rest_coords[0] - 360

            # sponge layers

            vs.rest_tscl = update(
                vs.rest_tscl,
                at[2:-2, 2:-2, :],
                veros.tools.interpolate(rest_coords, self._get_data(restoring_file, "tscl")[..., 0], t_grid),
            )

            t_star = self._get_data(restoring_file, "t_star")
            s_star = self._get_data(restoring_file, "s_star")
            for k in range(12):
                vs.t_star = update(
                    vs.t_star,
                    at[2:-2, 2:-2, :, k],
                    veros.tools.interpolate(rest_coords, t_star[..., k], t_grid, missing_value=0.0),
                )
                vs.s_star = update(
                    vs.s_star,
                    at[2:-2, 2:-2, :, k],
                    veros.tools.interpolate(rest_coords, s_star[..., k], t_grid, missing_value=0.0),
                )

    @veros_routine
    def set_forcing(self, state):
        vs = state.variables
        vs.update(set_forcing_kernel(state))

    @veros_routine
    def set_diagnostics(self, state):
        diagnostics = state.diagnostics
        settings = state.settings

        diagnostics["snapshot"].output_frequency = 3600.0 * 24 * 10
        diagnostics["averages"].output_frequency = 3600.0 * 24 * 10
        diagnostics["averages"].sampling_frequency = settings.dt_tracer
        diagnostics["averages"].output_variables = [
            "temp",
            "salt",
            "u",
            "v",
            "w",
            "surface_taux",
            "surface_tauy",
            "psi",
        ]
        diagnostics["cfl_monitor"].output_frequency = settings.dt_tracer * 10

    @veros_routine
    def after_timestep(self, state):
        pass


@veros_kernel
def set_forcing_kernel(state):
    vs = state.variables
    settings = state.settings

    year_in_seconds = 360 * 86400.0
    (n1, f1), (n2, f2) = veros.tools.get_periodic_interval(vs.time, year_in_seconds, year_in_seconds / 12.0, 12)

    vs.surface_taux = f1 * vs.taux[:, :, n1] + f2 * vs.taux[:, :, n2]
    vs.surface_tauy = f1 * vs.tauy[:, :, n1] + f2 * vs.tauy[:, :, n2]

    if settings.enable_tke:
        vs.forc_tke_surface = update(
            vs.forc_tke_surface,
            at[1:-1, 1:-1],
            npx.sqrt(
                (0.5 * (vs.surface_taux[1:-1, 1:-1] + vs.surface_taux[:-2, 1:-1]) / settings.rho_0) ** 2
                + (0.5 * (vs.surface_tauy[1:-1, 1:-1] + vs.surface_tauy[1:-1, :-2]) / settings.rho_0) ** 2
            )
            ** 1.5,
        )
    cp_0 = 3991.86795711963
    vs.forc_temp_surface = (
        (f1 * vs.sst_rest[:, :, n1] + f2 * vs.sst_rest[:, :, n2])
        * (f1 * vs.sst_clim[:, :, n1] + f2 * vs.sst_clim[:, :, n2] - vs.temp[:, :, -1, vs.tau])
        * vs.maskT[:, :, -1]
        / cp_0
        / settings.rho_0
    )
    vs.forc_salt_surface = (
        (f1 * vs.sss_rest[:, :, n1] + f2 * vs.sss_rest[:, :, n2])
        * (f1 * vs.sss_clim[:, :, n1] + f2 * vs.sss_clim[:, :, n2] - vs.salt[:, :, -1, vs.tau])
        * vs.maskT[:, :, -1]
    )

    ice_mask = (vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] <= -1.8) & (vs.forc_temp_surface <= 0.0)
    vs.forc_temp_surface = npx.where(ice_mask, 0.0, vs.forc_temp_surface)
    vs.forc_salt_surface = npx.where(ice_mask, 0.0, vs.forc_salt_surface)

    if settings.enable_tempsalt_sources:
        vs.temp_source = (
            vs.maskT
            * vs.rest_tscl
            * (f1 * vs.t_star[:, :, :, n1] + f2 * vs.t_star[:, :, :, n2] - vs.temp[:, :, :, vs.tau])
        )
        vs.salt_source = (
            vs.maskT
            * vs.rest_tscl
            * (f1 * vs.s_star[:, :, :, n1] + f2 * vs.s_star[:, :, :, n2] - vs.salt[:, :, :, vs.tau])
        )

    return KernelOutput(
        surface_taux=vs.surface_taux,
        surface_tauy=vs.surface_tauy,
        temp_source=vs.temp_source,
        salt_source=vs.salt_source,
        forc_tke_surface=vs.forc_tke_surface,
        forc_temp_surface=vs.forc_temp_surface,
        forc_salt_surface=vs.forc_salt_surface,
    )



from veros.core.external.streamfunction_init import streamfunction_init  # noqa: F401
from veros.core.external.solve_stream import solve_streamfunction  # noqa: F401
from veros.core.external.solve_pressure import solve_pressure  # noqa: F401



from veros.core.density.get_rho import (  # noqa: F401
    get_rho,
    get_potential_rho,
    get_dyn_enthalpy,
    get_salt,
    get_drhodT,
    get_drhodS,
    get_drhodp,
    get_int_drhodT,
    get_int_drhodS,
)



from abc import abstractmethod, ABCMeta


class LinearSolver(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, vs):
        pass

    @abstractmethod
    def solve(self, vs, rhs, x0, boundary_val=None):
        pass



from veros import logger
from veros.core.operators import numpy as npx
from veros.diagnostics.base import VerosDiagnostic
from veros.distributed import global_max


class CFLMonitor(VerosDiagnostic):
    """Diagnostic monitoring the maximum CFL number of the solution to detect
    instabilities.

    Writes output to stdout (no binary output).
    """

    name = "cfl_monitor"  #:
    output_frequency = None  #: Frequency (in seconds) in which output is written.

    def initialize(self, state):
        pass

    def diagnose(self, state):
        pass

    def output(self, state):
        """
        check for CFL violation
        """
        vs = state.variables
        settings = state.settings

        cfl = global_max(
            max(
                npx.max(
                    npx.abs(vs.u[2:-2, 2:-2, :, vs.tau])
                    * vs.maskU[2:-2, 2:-2, :]
                    / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
                    * settings.dt_tracer
                ),
                npx.max(
                    npx.abs(vs.v[2:-2, 2:-2, :, vs.tau])
                    * vs.maskV[2:-2, 2:-2, :]
                    / vs.dyt[npx.newaxis, 2:-2, npx.newaxis]
                    * settings.dt_tracer
                ),
            )
        )
        wcfl = global_max(
            npx.max(
                npx.abs(vs.w[2:-2, 2:-2, :, vs.tau])
                * vs.maskW[2:-2, 2:-2, :]
                / vs.dzt[npx.newaxis, npx.newaxis, :]
                * settings.dt_tracer
            )
        )

        if npx.isnan(cfl) or npx.isnan(wcfl):
            raise RuntimeError(f"CFL number is NaN at iteration {vs.itt}")

        logger.diagnostic(f" Maximal hor. CFL number = {cfl}")
        logger.diagnostic(f" Maximal ver. CFL number = {wcfl}")

        if settings.enable_eke or settings.enable_tke or settings.enable_idemix:
            cfl = global_max(
                max(
                    npx.max(
                        npx.abs(vs.u_wgrid[2:-2, 2:-2, :])
                        * vs.maskU[2:-2, 2:-2, :]
                        / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
                        * settings.dt_tracer
                    ),
                    npx.max(
                        npx.abs(vs.v_wgrid[2:-2, 2:-2, :])
                        * vs.maskV[2:-2, 2:-2, :]
                        / vs.dyt[npx.newaxis, 2:-2, npx.newaxis]
                        * settings.dt_tracer
                    ),
                )
            )
            wcfl = global_max(
                npx.max(
                    npx.abs(vs.w_wgrid[2:-2, 2:-2, :])
                    * vs.maskW[2:-2, 2:-2, :]
                    / vs.dzt[npx.newaxis, npx.newaxis, :]
                    * settings.dt_tracer
                )
            )
            logger.diagnostic(f" Maximal hor. CFL number on w grid = {cfl}")
            logger.diagnostic(f" Maximal ver. CFL number on w grid = {wcfl}")



import os
import h5netcdf

import veros.tools
from veros import VerosSetup, veros_routine, veros_kernel, KernelOutput, logger
from veros.variables import Variable
from veros.core.operators import numpy as npx, update, at

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_FILES = veros.tools.get_assets("global_4deg", os.path.join(BASE_PATH, "assets.json"))


class GlobalFourDegreeSetup(VerosSetup):
    """Global 4 degree model with 15 vertical levels.

    This setup demonstrates:
     - setting up a realistic model
     - reading input data from external files
     - including Indonesian throughflow
     - implementing surface forcings
     - applying a simple ice mask

    `Adapted from pyOM2 <https://wiki.cen.uni-hamburg.de/ifm/TO/pyOM2/4x4%20global%20model>`_.

    ChangeLog
     - 07-05-2020: modify bathymetry in order to include Indonesian throughflow;
       courtesy of Franka Jesse, Utrecht University

    """

    @veros_routine
    def set_parameter(self, state):
        settings = state.settings

        settings.identifier = "4deg"

        settings.nx, settings.ny, settings.nz = 90, 40, 15
        settings.dt_mom = 1800.0
        settings.dt_tracer = 86400.0
        settings.runlen = 0.0

        settings.x_origin = 4.0
        settings.y_origin = -76.0

        settings.coord_degree = True
        settings.enable_cyclic_x = True

        settings.enable_neutral_diffusion = True
        settings.K_iso_0 = 1000.0
        settings.K_iso_steep = 1000.0
        settings.iso_dslope = 4.0 / 1000.0
        settings.iso_slopec = 1.0 / 1000.0
        settings.enable_skew_diffusion = True

        settings.enable_hor_friction = True
        settings.A_h = (4 * settings.degtom) ** 3 * 2e-11
        settings.enable_hor_friction_cos_scaling = True
        settings.hor_friction_cosPower = 1

        settings.enable_implicit_vert_friction = True
        settings.enable_tke = True
        settings.c_k = 0.1
        settings.c_eps = 0.7
        settings.alpha_tke = 30.0
        settings.mxl_min = 1e-8
        settings.tke_mxl_choice = 2
        settings.kappaM_min = 2e-4
        settings.kappaH_min = 2e-5
        settings.enable_kappaH_profile = True
        settings.enable_tke_superbee_advection = True

        settings.enable_eke = True
        settings.eke_k_max = 1e4
        settings.eke_c_k = 0.4
        settings.eke_c_eps = 0.5
        settings.eke_cross = 2.0
        settings.eke_crhin = 1.0
        settings.eke_lmin = 100.0
        settings.enable_eke_superbee_advection = True

        settings.enable_idemix = False
        settings.enable_idemix_hor_diffusion = True
        settings.enable_eke_diss_surfbot = True
        settings.eke_diss_surfbot_frac = 0.2
        settings.enable_idemix_superbee_advection = True

        settings.eq_of_state_type = 5

        # custom variables
        state.dimensions["nmonths"] = 12
        state.var_meta.update(
            sss_clim=Variable("sss_clim", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            sst_clim=Variable("sst_clim", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            qnec=Variable("qnec", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            qnet=Variable("qnet", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            taux=Variable("taux", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            tauy=Variable("tauy", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
        )

    def _read_forcing(self, var):
        with h5netcdf.File(DATA_FILES["forcing"], "r") as infile:
            var_obj = infile.variables[var]
            return npx.array(var_obj).T

    @veros_routine
    def set_grid(self, state):
        vs = state.variables
        ddz = npx.array(
            [50.0, 70.0, 100.0, 140.0, 190.0, 240.0, 290.0, 340.0, 390.0, 440.0, 490.0, 540.0, 590.0, 640.0, 690.0]
        )
        vs.dzt = ddz[::-1]
        vs.dxt = 4.0 * npx.ones_like(vs.dxt)
        vs.dyt = 4.0 * npx.ones_like(vs.dyt)

    @veros_routine
    def set_coriolis(self, state):
        vs = state.variables
        settings = state.settings
        vs.coriolis_t = update(
            vs.coriolis_t, at[...], 2 * settings.omega * npx.sin(vs.yt[npx.newaxis, :] / 180.0 * settings.pi)
        )

    @veros_routine(dist_safe=False, local_variables=["kbot", "zt"])
    def set_topography(self, state):
        vs = state.variables
        settings = state.settings

        bathymetry_data = self._read_forcing("bathymetry")
        salt_data = self._read_forcing("salinity")[:, :, ::-1]
        land_mask = (vs.zt[npx.newaxis, npx.newaxis, :] <= bathymetry_data[..., npx.newaxis]) | (salt_data == 0.0)

        vs.kbot = update(vs.kbot, at[2:-2, 2:-2], 1 + npx.sum(land_mask.astype("int"), axis=2))

        # set all-land cells
        all_land_mask = (bathymetry_data == 0) | (vs.kbot[2:-2, 2:-2] == settings.nz)
        vs.kbot = update(vs.kbot, at[2:-2, 2:-2], npx.where(all_land_mask, 0, vs.kbot[2:-2, 2:-2]))

    @veros_routine(
        dist_safe=False,
        local_variables=[
            "taux",
            "tauy",
            "qnec",
            "qnet",
            "sss_clim",
            "sst_clim",
            "temp",
            "salt",
            "area_t",
            "maskT",
            "forc_iw_bottom",
            "forc_iw_surface",
        ],
    )
    def set_initial_conditions(self, state):
        vs = state.variables
        settings = state.settings

        # initial conditions for T and S
        temp_data = self._read_forcing("temperature")[:, :, ::-1]
        vs.temp = update(
            vs.temp, at[2:-2, 2:-2, :, :2], temp_data[:, :, :, npx.newaxis] * vs.maskT[2:-2, 2:-2, :, npx.newaxis]
        )

        salt_data = self._read_forcing("salinity")[:, :, ::-1]
        vs.salt = update(
            vs.salt, at[2:-2, 2:-2, :, :2], salt_data[..., npx.newaxis] * vs.maskT[2:-2, 2:-2, :, npx.newaxis]
        )

        # use Trenberth wind stress from MITgcm instead of ECMWF (also contained in ecmwf_4deg.cdf)
        vs.taux = update(vs.taux, at[2:-2, 2:-2, :], self._read_forcing("tau_x"))
        vs.tauy = update(vs.tauy, at[2:-2, 2:-2, :], self._read_forcing("tau_y"))

        # heat flux
        with h5netcdf.File(DATA_FILES["ecmwf"], "r") as ecmwf_data:
            qnec_var = ecmwf_data.variables["Q3"]
            vs.qnec = update(vs.qnec, at[2:-2, 2:-2, :], npx.array(qnec_var).T)
            vs.qnec = npx.where(vs.qnec <= -1e10, 0.0, vs.qnec)

        q = self._read_forcing("q_net")
        vs.qnet = update(vs.qnet, at[2:-2, 2:-2, :], -q)
        vs.qnet = npx.where(vs.qnet <= -1e10, 0.0, vs.qnet)

        mean_flux = (
            npx.sum(vs.qnet[2:-2, 2:-2, :] * vs.area_t[2:-2, 2:-2, npx.newaxis]) / 12 / npx.sum(vs.area_t[2:-2, 2:-2])
        )
        logger.info(" removing an annual mean heat flux imbalance of %e W/m^2" % mean_flux)
        vs.qnet = (vs.qnet - mean_flux) * vs.maskT[:, :, -1, npx.newaxis]

        # SST and SSS
        vs.sst_clim = update(vs.sst_clim, at[2:-2, 2:-2, :], self._read_forcing("sst"))
        vs.sss_clim = update(vs.sss_clim, at[2:-2, 2:-2, :], self._read_forcing("sss"))

        if settings.enable_idemix:
            vs.forc_iw_bottom = update(
                vs.forc_iw_bottom, at[2:-2, 2:-2], self._read_forcing("tidal_energy") / settings.rho_0
            )
            vs.forc_iw_surface = update(
                vs.forc_iw_surface, at[2:-2, 2:-2], self._read_forcing("wind_energy") / settings.rho_0 * 0.2
            )

    @veros_routine
    def set_forcing(self, state):
        vs = state.variables
        vs.update(set_forcing_kernel(state))

    @veros_routine
    def set_diagnostics(self, state):
        settings = state.settings
        state.diagnostics["snapshot"].output_frequency = 360 * 86400.0
        state.diagnostics["overturning"].output_frequency = 360 * 86400.0
        state.diagnostics["overturning"].sampling_frequency = settings.dt_tracer
        state.diagnostics["energy"].output_frequency = 360 * 86400.0
        state.diagnostics["energy"].sampling_frequency = 86400
        average_vars = ["temp", "salt", "u", "v", "w", "surface_taux", "surface_tauy", "psi"]
        state.diagnostics["averages"].output_variables = average_vars
        state.diagnostics["averages"].output_frequency = 360 * 86400.0
        state.diagnostics["averages"].sampling_frequency = 86400

    @veros_routine
    def after_timestep(self, state):
        pass


@veros_kernel
def set_forcing_kernel(state):
    vs = state.variables
    settings = state.settings

    year_in_seconds = 360 * 86400.0
    (n1, f1), (n2, f2) = veros.tools.get_periodic_interval(vs.time, year_in_seconds, year_in_seconds / 12.0, 12)

    # wind stress
    vs.surface_taux = f1 * vs.taux[:, :, n1] + f2 * vs.taux[:, :, n2]
    vs.surface_tauy = f1 * vs.tauy[:, :, n1] + f2 * vs.tauy[:, :, n2]

    # tke flux
    if settings.enable_tke:
        vs.forc_tke_surface = update(
            vs.forc_tke_surface,
            at[1:-1, 1:-1],
            npx.sqrt(
                (0.5 * (vs.surface_taux[1:-1, 1:-1] + vs.surface_taux[:-2, 1:-1]) / settings.rho_0) ** 2
                + (0.5 * (vs.surface_tauy[1:-1, 1:-1] + vs.surface_tauy[1:-1, :-2]) / settings.rho_0) ** 2
            )
            ** 1.5,
        )
    # heat flux : W/m^2 K kg/J m^3/kg = K m/s
    cp_0 = 3991.86795711963
    sst = f1 * vs.sst_clim[:, :, n1] + f2 * vs.sst_clim[:, :, n2]
    qnec = f1 * vs.qnec[:, :, n1] + f2 * vs.qnec[:, :, n2]
    qnet = f1 * vs.qnet[:, :, n1] + f2 * vs.qnet[:, :, n2]
    vs.forc_temp_surface = (
        (qnet + qnec * (sst - vs.temp[:, :, -1, vs.tau])) * vs.maskT[:, :, -1] / cp_0 / settings.rho_0
    )

    # salinity restoring
    t_rest = 30 * 86400.0
    sss = f1 * vs.sss_clim[:, :, n1] + f2 * vs.sss_clim[:, :, n2]
    vs.forc_salt_surface = 1.0 / t_rest * (sss - vs.salt[:, :, -1, vs.tau]) * vs.maskT[:, :, -1] * vs.dzt[-1]

    # apply simple ice mask
    mask = npx.logical_and(vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] < -1.8, vs.forc_temp_surface < 0.0)
    vs.forc_temp_surface = npx.where(mask, 0.0, vs.forc_temp_surface)
    vs.forc_salt_surface = npx.where(mask, 0.0, vs.forc_salt_surface)

    return KernelOutput(
        surface_taux=vs.surface_taux,
        surface_tauy=vs.surface_tauy,
        forc_tke_surface=vs.forc_tke_surface,
        forc_temp_surface=vs.forc_temp_surface,
        forc_salt_surface=vs.forc_salt_surface,
    )
